# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import os
import time

from functools import wraps

from prometheus_client import Gauge, Histogram, Counter


def decode_buckets(buckets_list):
    return [float(x) for x in buckets_list.split(",")]


def get_histogram_buckets_from_evn(env_name):
    if env_name in os.environ:
        buckets = decode_buckets(os.environ.get(env_name))
    else:
        if hasattr(Histogram, "DEFAULT_BUCKETS"):  # pragma: no cover
            buckets = Histogram.DEFAULT_BUCKETS
        else:  # pragma: no cover
            # For prometheus-client < 0.3.0 we cannot easily access
            # the default buckets:
            buckets = (
                0.005,
                0.01,
                0.025,
                0.05,
                0.075,
                0.1,
                0.25,
                0.5,
                0.75,
                1.0,
                2.5,
                5.0,
                7.5,
                10.0,
                float("inf"),
            )
    return buckets


def setup_gauge(*gauges):
    def wrapper(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            for g in gauges:
                g.inc(1)
            try:
                return func(*args, **kwargs)
            finally:
                for g in gauges:
                    g.dec(1)

        return _wrapper

    return wrapper


def setup_histogram(*histograms):
    def wrapper(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            start = time.time()
            try:
                return func(*args, **kwargs)
            finally:
                for h in histograms:
                    h.observe(time.time() - start)

        return _wrapper

    return wrapper


# engine metrics
ENGINE_RUNNING_PROCESSES = Gauge(
    "engine_running_processes", "count running state processes"
)
ENGINE_RUNNING_SCHEDULES = Gauge(
    "engine_running_schedules", "count running state schedules"
)
ENGINE_PROCESS_RUNNING_TIME = Histogram(
    "engine_process_running_time",
    "time spent running process",
    buckets=get_histogram_buckets_from_evn("ENGINE_PROCESS_RUNNING_TIME_BUCKETS"),
)
ENGINE_SCHEDULE_RUNNING_TIME = Histogram(
    "engine_schedule_running_time",
    "time spent running schedule",
    buckets=get_histogram_buckets_from_evn("ENGINE_SCHEDULE_RUNNING_TIME_BUCKETS"),
)
ENGINE_NODE_EXECUTE_TIME = Histogram(
    "engine_node_execute_time",
    "time spent executing node",
    buckets=get_histogram_buckets_from_evn("ENGINE_NODE_EXECUTE_TIME_BUCKETS"),
    labelnames=["type"],
)
ENGINE_NODE_SCHEDULE_TIME = Histogram(
    "engine_node_schedule_time",
    "time spent scheduling node",
    buckets=get_histogram_buckets_from_evn("ENGINE_NODE_SCHEDULE_TIME_BUCKETS"),
    labelnames=["type"],
)

# runtime metrics
ENGINE_RUNTIME_CONTEXT_VALUE_READ_TIME = Histogram(
    "engine_runtime_context_value_read_time", "time spent reading context value"
)
ENGINE_RUNTIME_CONTEXT_REF_READ_TIME = Histogram(
    "engine_runtime_context_ref_read_time", "time spent reading context value reference"
)
ENGINE_RUNTIME_CONTEXT_VALUE_UPSERT_TIME = Histogram(
    "engine_runtime_context_value_upsert_time", "time spent upserting context value"
)

ENGINE_RUNTIME_DATA_INPUTS_READ_TIME = Histogram(
    "engine_runtime_data_inputs_read_time", "time spent reading node data inputs"
)
ENGINE_RUNTIME_DATA_OUTPUTS_READ_TIME = Histogram(
    "engine_runtime_data_outputs_read_time", "time spent reading node data outputs"
)
ENGINE_RUNTIME_DATA_READ_TIME = Histogram(
    "engine_runtime_data_read_time", "time spent reading node data inputs and outputs"
)

ENGINE_RUNTIME_EXEC_DATA_INPUTS_READ_TIME = Histogram(
    "engine_runtime_exec_data_inputs_read_time",
    "time spent reading node execution data inputs",
)
ENGINE_RUNTIME_EXEC_DATA_OUTPUTS_READ_TIME = Histogram(
    "engine_runtime_exec_data_outputs_read_time",
    "time spent reading node execution data outputs",
)
ENGINE_RUNTIME_EXEC_DATA_READ_TIME = Histogram(
    "engine_runtime_exec_data_read_time",
    "time spent reading node execution data inputs and outputs",
)
ENGINE_RUNTIME_EXEC_DATA_INPUTS_WRITE_TIME = Histogram(
    "engine_runtime_exec_data_inputs_write_time",
    "time spent writing node execution data inputs",
)
ENGINE_RUNTIME_EXEC_DATA_OUTPUTS_WRITE_TIME = Histogram(
    "engine_runtime_exec_data_outputs_write_time",
    "time spent writing node execution data outputs",
)
ENGINE_RUNTIME_EXEC_DATA_WRITE_TIME = Histogram(
    "engine_runtime_exec_data_write_time",
    "time spent writing node execution data inputs and outputs",
)
ENGINE_RUNTIME_CALLBACK_DATA_READ_TIME = Histogram(
    "engine_runtime_callback_data_read_time",
    "time spent reading node callback data",
)

ENGINE_RUNTIME_SCHEDULE_READ_TIME = Histogram(
    "engine_runtime_schedule_read_time", "time spent reading schedule"
)
ENGINE_RUNTIME_SCHEDULE_WRITE_TIME = Histogram(
    "engine_runtime_schedule_write_time", "time spent writing schedule"
)

ENGINE_RUNTIME_STATE_READ_TIME = Histogram(
    "engine_runtime_state_read_time", "time spent reading state"
)
ENGINE_RUNTIME_STATE_WRITE_TIME = Histogram(
    "engine_runtime_state_write_time", "time spent writing state"
)

ENGINE_RUNTIME_NODE_READ_TIME = Histogram(
    "engine_runtime_node_read_time", "time spent reading node"
)

ENGINE_RUNTIME_PROCESS_READ_TIME = Histogram(
    "engine_runtime_process_read_time", "time spent reading process"
)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
ERI 中相关的模型对象
"""

from enum import Enum
from datetime import datetime
from typing import List, Dict, Any, Optional
from abc import ABCMeta, abstractmethod

from bamboo_engine.utils.object import Representable
from bamboo_engine.utils.collections import FancyDict
from bamboo_engine.exceptions import ValueError


# node relate models
class NodeType(Enum):
    """
    节点类型枚举
    """

    ServiceActivity = "ServiceActivity"
    SubProcess = "SubProcess"
    ExclusiveGateway = "ExclusiveGateway"
    ParallelGateway = "ParallelGateway"
    ConditionalParallelGateway = "ConditionalParallelGateway"
    ConvergeGateway = "ConvergeGateway"
    EmptyStartEvent = "EmptyStartEvent"
    EmptyEndEvent = "EmptyEndEvent"
    ExecutableEndEvent = "ExecutableEndEvent"


class Node(Representable):
    """
    节点信息描述类
    """

    def __init__(
        self,
        id: str,
        type: NodeType,
        target_flows: List[str],
        target_nodes: List[str],
        targets: Dict[str, str],
        root_pipeline_id: str,
        parent_pipeline_id: str,
        can_skip: bool = True,
        can_retry: bool = True,
    ):
        """

        :param id: 节点 ID
        :type id: str
        :param type: 节点类型
        :type type: NodeType
        :param target_flows: 节点目标流 ID 列表
        :type target_flows: List[str]
        :param target_nodes: 目标节点 ID 列表
        :type target_nodes: List[str]
        :param targets: 节点目标流，目标节点 ID 映射
        :type targets: Dict[str, str]
        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param parent_pipeline_id: 父流程  ID
        :type parent_pipeline_id: str
        :param can_skip: 节点是否能够跳过
        :type can_skip: bool
        :param can_retry: 节点是否能够重试
        :type can_retry: bool
        """
        self.id = id
        self.type = type
        self.targets = targets
        self.target_flows = target_flows
        self.target_nodes = target_nodes
        self.root_pipeline_id = root_pipeline_id
        self.parent_pipeline_id = parent_pipeline_id
        self.can_skip = can_skip
        self.can_retry = can_retry


class EmptyStartEvent(Node):
    pass


class ConvergeGateway(Node):
    pass


class EmptyEndEvent(Node):
    pass


class Condition(Representable):
    """
    分支条件
    """

    def __init__(self, name: str, evaluation: str, target_id: str, flow_id: str):
        """

        :param name: 条件名
        :type name: str
        :param evaluation: 条件表达式
        :type evaluation: str
        :param target_id: 目标节点 ID
        :type target_id: str
        :param flow_id: 目标流 ID
        :type flow_id: str
        """
        self.name = name
        self.evaluation = evaluation
        self.target_id = target_id
        self.flow_id = flow_id


class ParallelGateway(Node):
    """
    并行网关
    """

    def __init__(self, converge_gateway_id: str, *args, **kwargs):
        """

        :param converge_gateway_id: 汇聚网关 ID
        :type converge_gateway_id: str
        """
        super().__init__(*args, **kwargs)
        self.converge_gateway_id = converge_gateway_id


class ConditionalParallelGateway(Node):
    """
    条件并行网关
    """

    def __init__(self, conditions: List[Condition], converge_gateway_id: str, *args, **kwargs):
        """

        :param conditions: 分支条件
        :type conditions: List[Condition]
        :param converge_gateway_id: 汇聚网关 ID
        :type converge_gateway_id: str
        """
        super().__init__(*args, **kwargs)
        self.conditions = conditions
        self.converge_gateway_id = converge_gateway_id


class ExclusiveGateway(Node):
    """
    分支网关
    """

    def __init__(self, conditions: List[Condition], *args, **kwargs):
        """

        :param conditions: 分支条件
        :type conditions: List[Condition]
        """
        super().__init__(*args, **kwargs)
        self.conditions = conditions


class ServiceActivity(Node):
    """
    服务节点
    """

    def __init__(self, code: str, version: str, timeout: Optional[int], error_ignorable: bool, *args, **kwargs):
        """

        :param code: Service Code
        :type code: str
        :param version: 版本
        :type version: str
        :param timeout: 超时限制
        :type timeout: Optional[int]
        :param error_ignorable: 是否忽略错误
        :type error_ignorable: bool
        """

        super().__init__(*args, **kwargs)
        self.code = code
        self.version = version
        self.timeout = timeout
        self.error_ignorable = error_ignorable


class SubProcess(Node):
    """
    子流程
    """

    def __init__(self, start_event_id: str, *args, **kwargs):
        """

        :param start_event_id: 子流程开始节点 ID
        :type start_event_id: str
        """
        super().__init__(*args, **kwargs)
        self.start_event_id = start_event_id


class ExecutableEndEvent(Node):
    """
    可执行结束节点
    """

    def __init__(self, code: str, *args, **kwargs):
        """

        :param code: 可执行结束节点 ID
        :type code: str
        """
        super().__init__(*args, **kwargs)
        self.code = code


# runtime relate models
class ScheduleType(Enum):
    """
    调度类型
    """

    CALLBACK = 1
    MULTIPLE_CALLBACK = 2
    POLL = 3


class Schedule(Representable):
    """
    调度对象
    """

    def __init__(
        self,
        id: int,
        type: ScheduleType,
        process_id: int,
        node_id: str,
        finished: bool,
        expired: bool,
        version: str,
        times: int,
    ):
        """

        :param id: ID
        :type id: int
        :param type: 类型
        :type type: ScheduleType
        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param finished: 是否已完成
        :type finished: bool
        :param expired: 是否已过期
        :type expired: bool
        :param version: 绑定版本
        :type version: str
        :param times: 调度次数
        :type times: int
        """
        self.id = id
        self.type = type
        self.process_id = process_id
        self.node_id = node_id
        self.finished = finished
        self.expired = expired
        self.version = version
        self.times = times


class State(Representable):
    """
    节点状态对象
    """

    def __init__(
        self,
        node_id: str,
        root_id: str,
        parent_id: str,
        name: str,
        version: str,
        loop: int,
        retry: int,
        skip: bool,
        error_ignored: bool,
        created_time: datetime,
        started_time: datetime,
        archived_time: datetime,
    ):
        """
        :param node_id: 节点 ID
        :type node_id: str
        :param root_id: 根流程 ID
        :type root_id: str
        :param parent_id: 父流程 ID
        :type parent_id: str
        :param name: 状态名
        :type name: str
        :param version: 版本
        :type version: str
        :param loop: 重入次数
        :type loop: int
        :param retry: 重试次数
        :type retry: int
        :param skip: 是否跳过
        :type skip: bool
        :param error_ignored: 是否出错后自动忽略
        :type error_ignored: bool
        :param started_time: 创建时间
        :type started_time: datetime
        :param started_time: 开始时间
        :type started_time: datetime
        :param archived_time: 归档时间
        :type archived_time: datetime
        """
        self.node_id = node_id
        self.root_id = root_id
        self.parent_id = parent_id
        self.name = name
        self.version = version
        self.loop = loop
        self.retry = retry
        self.skip = skip
        self.error_ignored = error_ignored
        self.created_time = created_time
        self.started_time = started_time
        self.archived_time = archived_time


class DataInput(Representable):
    """
    节点数据输入项
    """

    def __init__(self, need_render: bool, value: Any):
        """
        :type is_splice: bool
        :param value: 是否需要进行模板解析
        :type value: Any
        """
        self.need_render = need_render
        self.value = value


class Data(Representable):
    """
    节点数据对象
    """

    def __init__(self, inputs: Dict[str, DataInput], outputs: Dict[str, str]):
        """

        :param inputs: 输入数据
        :type inputs: Dict[str, Any]
        :param outputs: 节点输出配置
        :type outputs: Dict[str, str]
        """
        self.inputs = inputs
        self.outputs = outputs

    def plain_inputs(self) -> Dict[str, Any]:
        """
        获取不带输入项类型的输入字典
        """
        return {key: di.value for key, di in self.inputs.items()}

    def need_render_inputs(self) -> Dict[str, Any]:
        """
        获取需要进行渲染的输入项字典
        """
        return {key: di.value for key, di in self.inputs.items() if di.need_render}

    def render_escape_inputs(self) -> Dict[str, Any]:
        """
        获取不需要进行渲染的输入项字典
        """
        return {key: di.value for key, di in self.inputs.items() if not di.need_render}


class ExecutionData(Representable):
    """
    节点输出数据
    """

    def __init__(self, inputs: Optional[dict], outputs: Optional[dict]):
        """

        :param inputs: 输入数据
        :type inputs: Optional[dict]
        :param outputs: 输出数据
        :type outputs: Optional[dict]
        """
        self.inputs = FancyDict(inputs)
        self.outputs = FancyDict(outputs)


class ExecutionHistory(Representable):
    """
    节点执行历史
    """

    def __init__(
        self,
        id: str,
        node_id: str,
        started_time: datetime,
        archived_time: datetime,
        loop: int,
        skip: bool,
        retry: int,
        version: str,
        inputs: dict,
        outputs: dict,
    ):
        """

        : param id: ID
        : type id: str
        : param node_id: Node ID
        : type node_id: str
        : param started_time: 开始时间
        : type started_time: datetime
        : param archived_time: 归档时间
        : type archived_time: datetime
        : param loop: 重入计数
        : type loop: int
        : param skip: 是否跳过
        : type skip: bool
        : param retry: 重试次数
        : type retry: int
        : param version: 版本号
        : type version: str
        : param inputs: 输入数据
        : type inputs: dict
        : param outputs: 输出数据
        : type outputs: dict
        """
        self.id = id
        self.node_id = node_id
        self.started_time = started_time
        self.archived_time = archived_time
        self.loop = loop
        self.skip = skip
        self.retry = retry
        self.version = version
        self.inputs = inputs
        self.outputs = outputs


class ExecutionShortHistory(Representable):
    """
    简短节点执行历史
    """

    def __init__(
        self,
        id: str,
        node_id: str,
        started_time: datetime,
        archived_time: datetime,
        loop: int,
        skip: bool,
        retry: int,
        version: str,
    ):
        """

        : param id: ID
        : type id: str
        : param node_id: Node ID
        : type node_id: str
        : param started_time: 开始时间
        : type started_time: datetime
        : param archived_time: 归档时间
        : type archived_time: datetime
        : param loop: 重入计数
        : type loop: int
        : param skip: 是否跳过
        : type skip: bool
        : param retry: 重试次数
        : type retry: int
        : param version: 版本号
        : type version: str
        """
        self.id = id
        self.node_id = node_id
        self.started_time = started_time
        self.archived_time = archived_time
        self.loop = loop
        self.skip = skip
        self.retry = retry
        self.version = version


class CallbackData(Representable):
    """
    节点回调数据
    """

    def __init__(self, id: int, node_id: str, version: str, data: dict):
        """

        :param id: 数据 ID
        :type id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param version: 版本
        :type version: str
        :param data: 数据
        :type data: dict
        """
        self.id = id
        self.node_id = node_id
        self.version = version
        self.data = data


class SuspendedProcessInfo(Representable):
    """
    挂起进程信息
    """

    def __init__(self, process_id: int, current_node: str):
        """

        :param process_id: 进程 ID
        :type process_id: int
        :param current_node: 当前节点 ID
        :type current_node: str
        """
        self.process_id = process_id
        self.current_node = current_node


class ProcessInfo(Representable):
    """
    进程信息
    """

    def __init__(
        self,
        process_id: int,
        destination_id: str,
        root_pipeline_id: str,
        pipeline_stack: List[str],
        parent_id: int,
    ):
        """

        :param process_id: 进程 ID
        :type process_id: int
        :param destination_id: 进程目标节点 ID
        :type destination_id: str
        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param pipeline_stack: 流程栈
        :type pipeline_stack: List[str]
        :param parent_id: 父进程 ID
        :type parent_id: int
        """
        self.process_id = process_id
        self.destination_id = destination_id
        self.parent_id = parent_id
        self.root_pipeline_id = root_pipeline_id
        self.pipeline_stack = pipeline_stack

    @property
    def top_pipeline_id(self):
        return self.pipeline_stack[-1]


class DispatchProcess(Representable):
    """
    待调度进程信息
    """

    def __init__(self, process_id: int, node_id: str):
        """

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 调度开始节点 ID
        :type node_id: str
        """
        self.process_id = process_id
        self.node_id = node_id


class ContextValueType(Enum):
    """

    :param Enum: [description]
    :type Enum: [type]
    """

    PLAIN = 1
    SPLICE = 2
    COMPUTE = 3


class ContextValue(Representable):
    def __init__(self, key: str, type: ContextValueType, value: Any, code: Optional[str] = None):
        if type is ContextValueType.COMPUTE and code is None:
            raise ValueError("code can't be none when type is COMPUTE")

        self.key = key
        self.type = type
        self.value = value
        self.code = code



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
import logging

from bamboo_engine import states
from bamboo_engine.context import Context
from bamboo_engine.template import Template
from bamboo_engine.handler import register_handler, NodeHandler, ExecuteResult
from bamboo_engine.utils.boolrule import BoolRule
from bamboo_engine.eri import NodeType, ProcessInfo

from bamboo_engine.utils.string import transform_escape_char

logger = logging.getLogger("bamboo_engine")


@register_handler(NodeType.ExclusiveGateway)
class ExclusiveGatewayHandler(NodeHandler):
    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :return: 执行结果
        :rtype: ExecuteResult
        """
        evaluations = [c.evaluation for c in self.node.conditions]
        top_pipeline_id = process_info.top_pipeline_id
        root_pipeline_id = process_info.root_pipeline_id

        root_pipeline_inputs = self._get_plain_inputs(process_info.root_pipeline_id)

        # resolve conditions references
        evaluation_refs = set()
        for e in evaluations:
            refs = Template(e).get_reference()
            evaluation_refs = evaluation_refs.union(refs)

        logger.info(
            "[%s] %s evaluation original refs: %s",
            root_pipeline_id,
            self.node.id,
            evaluation_refs,
        )
        additional_refs = self.runtime.get_context_key_references(
            pipeline_id=top_pipeline_id, keys=evaluation_refs
        )
        evaluation_refs = evaluation_refs.union(additional_refs)

        logger.info(
            "[%s] %s evaluation final refs: %s",
            root_pipeline_id,
            self.node.id,
            evaluation_refs,
        )
        context_values = self.runtime.get_context_values(
            pipeline_id=top_pipeline_id, keys=evaluation_refs
        )
        logger.info(
            "[%s] %s evaluation context values: %s",
            root_pipeline_id,
            self.node.id,
            context_values,
        )

        context = Context(self.runtime, context_values, root_pipeline_inputs)
        hydrated_context = {
            k: transform_escape_char(v)
            for k, v in context.hydrate(deformat=True).items()
        }

        # check conditions
        meet_targets = []
        meet_conditions = []
        for c in self.node.conditions:
            resolved_evaluate = Template(c.evaluation).render(hydrated_context)
            logger.info(
                "[%s] %s render evaluation %s: %s with %s",
                root_pipeline_id,
                self.node.id,
                c.evaluation,
                resolved_evaluate,
                hydrated_context,
            )
            try:
                result = BoolRule(resolved_evaluate).test()
            except Exception as e:
                # test failed
                return self._execute_fail(
                    "evaluate[{}] fail with data[{}] message: {}".format(
                        resolved_evaluate, json.dumps(hydrated_context), e
                    )
                )
            else:
                if result:
                    meet_conditions.append(c.name)
                    meet_targets.append(c.target_id)

        # all miss
        if not meet_targets:
            return self._execute_fail("all conditions of branches are not meet")

        # multiple branch hit
        if len(meet_targets) != 1:
            return self._execute_fail(
                "multiple conditions meet: {}".format(meet_conditions)
            )

        self.runtime.set_state(
            node_id=self.node.id, to_state=states.FINISHED, set_archive_time=True
        )

        return ExecuteResult(
            should_sleep=False,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=[],
            next_node_id=meet_targets[0],
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
字符串处理类工具
"""

import uuid

ESCAPED_CHARS = {"\n": r"\n", "\r": r"\r", "\t": r"\t"}


def transform_escape_char(string: str) -> str:
    """
    对未转义的字符串进行转义，现有的转义字符包括\n, \r, \t
    """
    if not isinstance(string, str):
        return string
    # 已转义的情况
    if len([c for c in ESCAPED_CHARS.values() if c in string]) > 0:
        return string
    for key, value in ESCAPED_CHARS.items():
        if key in string:
            string = string.replace(key, value)
    return string


def format_var_key(key: str) -> str:
    """
    format key to ${key}

    :param key: key
    :type key: str
    :return: format key
    :rtype: str
    """
    return "${%s}" % key


def deformat_var_key(key: str) -> str:
    """
    deformat ${key} to key

    :param key: key
    :type key: str
    :return: deformat key
    :rtype: str
    """
    return key[2:-1]


def unique_id(prefix: str) -> str:
    if len(prefix) != 1:
        raise ValueError("prefix length must be 1")

    return "{}{}".format(prefix, uuid.uuid4().hex)


def get_lower_case_name(text: str) -> str:
    lst = []
    for index, char in enumerate(text):
        if char.isupper() and index != 0:
            lst.append("_")
        lst.append(char)

    return "".join(lst).lower()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from bamboo_engine import states
from bamboo_engine.context import Context
from bamboo_engine.config import Settings
from bamboo_engine.template import Template
from bamboo_engine.eri import ProcessInfo, ContextValue, ContextValueType, NodeType
from bamboo_engine.handler import register_handler, NodeHandler, ExecuteResult
from bamboo_engine.utils.string import format_var_key

logger = logging.getLogger("bamboo_engine")


@register_handler(NodeType.SubProcess)
class SubProcessHandler(NodeHandler):
    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :return: 执行结果
        :rtype: ExecuteResult
        """
        data = self.runtime.get_data(self.node.id)
        root_pipeline_inputs = self._get_plain_inputs(process_info.root_pipeline_id)
        need_render_inputs = data.need_render_inputs()
        render_escape_inputs = data.render_escape_inputs()
        top_pipeline_id = process_info.top_pipeline_id
        root_pipeline_id = process_info.root_pipeline_id

        logger.info(
            "[%s] %s subprocess data: %s",
            root_pipeline_id,
            self.node.id,
            data,
        )

        # resolve inputs context references
        inputs_refs = Template(need_render_inputs).get_reference()
        logger.info(
            "[%s] %s subprocess original refs: %s",
            root_pipeline_id,
            self.node.id,
            inputs_refs,
        )

        additional_refs = self.runtime.get_context_key_references(
            pipeline_id=top_pipeline_id, keys=inputs_refs
        )
        inputs_refs = inputs_refs.union(additional_refs)
        logger.info(
            "[%s] %s subprocess final refs: %s",
            root_pipeline_id,
            self.node.id,
            inputs_refs,
        )

        # prepare context
        context_values = self.runtime.get_context_values(
            pipeline_id=top_pipeline_id, keys=inputs_refs
        )

        # pre extract loop outputs
        loop_value = loop + Settings.RERUN_INDEX_OFFSET
        if self.LOOP_KEY in data.outputs:
            loop_output_key = data.outputs[self.LOOP_KEY]
            context_values.append(
                ContextValue(
                    key=loop_output_key,
                    type=ContextValueType.PLAIN,
                    value=loop_value,
                )
            )
        logger.info(
            "[%s] %s subprocess parent context values: %s",
            root_pipeline_id,
            self.node.id,
            context_values,
        )

        context = Context(self.runtime, context_values, root_pipeline_inputs)
        hydrated_context = context.hydrate(deformat=True)
        logger.info(
            "[%s] %s subprocess parent hydrated context: %s",
            root_pipeline_id,
            self.node.id,
            hydrated_context,
        )

        # resolve inputs
        subprocess_inputs = Template(need_render_inputs).render(hydrated_context)
        subprocess_inputs.update(render_escape_inputs)
        sub_context_values = {
            key: ContextValue(key=key, type=ContextValueType.PLAIN, value=value)
            for key, value in subprocess_inputs.items()
        }
        logger.info(
            "[%s] %s subprocess inject context: %s",
            root_pipeline_id,
            self.node.id,
            sub_context_values,
        )

        # update subprocess context, inject subprocess data
        self.runtime.upsert_plain_context_values(self.node.id, sub_context_values)
        process_info.pipeline_stack.append(self.node.id)
        self.runtime.set_pipeline_stack(
            process_info.process_id, process_info.pipeline_stack
        )

        return ExecuteResult(
            should_sleep=False,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=[],
            next_node_id=self.node.start_event_id,
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from bamboo_engine import states
from bamboo_engine.eri import ProcessInfo, NodeType
from bamboo_engine.handler import register_handler, NodeHandler, ExecuteResult


@register_handler(NodeType.ParallelGateway)
class ParallelGatewayHandler(NodeHandler):
    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :return: 执行结果
        :rtype: ExecuteResult
        """

        from_to = {}
        for target in self.node.target_nodes:
            from_to[target] = self.node.converge_gateway_id

        dispatch_processes = self.runtime.fork(
            parent_id=process_info.process_id,
            root_pipeline_id=process_info.root_pipeline_id,
            pipeline_stack=process_info.pipeline_stack,
            from_to=from_to,
        )

        self.runtime.set_state(
            node_id=self.node.id, to_state=states.FINISHED, set_archive_time=True
        )

        return ExecuteResult(
            should_sleep=True,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=dispatch_processes,
            next_node_id=None,
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
引擎执行 local
"""

from typing import Optional

from werkzeug.local import Local

from .utils.object import Representable

_local = Local()


class CurrentNodeInfo(Representable):
    def __init__(self, node_id: str, version: str, loop: int):
        self.node_id = node_id
        self.version = version
        self.loop = loop


def set_node_info(node_info: CurrentNodeInfo):
    """
    设置当前进程/线程/协程 Local 中的当前节点信息

    :param node_id: 节点 ID
    :type node_id: str
    :param version: 节点版本
    :type version: str
    :param loop: 重入次数
    :type loop: int
    """
    _local.current_node_info = node_info


def get_node_info() -> Optional[CurrentNodeInfo]:
    """
    获取当前进程/线程/协程正在处理的节点 ID，版本及重入次数

    :return: 节点 ID
    :rtype: [type]
    """
    return getattr(_local, "current_node_info", None)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
流程上下文相关逻辑封装模块
"""

from weakref import WeakValueDictionary
from typing import List, Dict, Any

from bamboo_engine.eri import (
    ContextValue,
    ExecutionData,
    EngineRuntimeInterface,
    Variable,
    ContextValueType,
)
from .template.template import Template
from .utils.string import format_var_key, deformat_var_key


class PlainVariable(Variable):
    """
    普通变量
    """

    def __init__(self, key: str, value: Any):
        self.key = key
        self.value = value

    def get(self):
        return self.value


class SpliceVariable(Variable):
    """
    模板类型变量，会尝试在流程上下文中解析变量中定义的模板
    """

    def __init__(self, key: str, value: Any, pool: WeakValueDictionary):
        self.key = key
        self.value = value
        self.pool = pool
        self.refs = [k for k in Template(value).get_reference()]
        import pdb

    def get(self):
        context = {}
        for r in self.refs:
            if r not in self.pool:
                continue

            var = self.pool[r]
            if issubclass(var.__class__, Variable):
                var = var.get()
            context[deformat_var_key(r)] = var

        return Template(self.value).render(context=context)


def _raw_key(key: str) -> str:
    return key


class Context:
    """
    流程执行上下文，封装引擎在执行流程的过程中对上下文进行的操作和逻辑
    """

    def __init__(
        self,
        runtime: EngineRuntimeInterface,
        values: List[ContextValue],
        additional_data: dict,
    ):
        """

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param values: 上下文数据列表
        :type values: List[ContextValue]
        :param additional_data: 额外数据字典
        :type additional_data: dict
        """
        self.values = values
        self.runtime = runtime
        self.pool = WeakValueDictionary()
        self.variables = {}
        self.additional_data = additional_data

        # 将上下文数据转换成变量，变量内封装了自身解析的逻辑，且实现了 Variable 接口
        for v in self.values:
            if v.type is ContextValueType.PLAIN:
                self.variables[v.key] = PlainVariable(key=v.key, value=v.value)
            elif v.type is ContextValueType.SPLICE:
                self.variables[v.key] = SpliceVariable(
                    key=v.key, value=v.value, pool=self.pool
                )
            elif v.type is ContextValueType.COMPUTE:
                self.variables[v.key] = self.runtime.get_compute_variable(
                    code=v.code,
                    key=v.key,
                    value=SpliceVariable(key=v.key, value=v.value, pool=self.pool),
                    additional_data=self.additional_data,
                )

        for k, var in self.variables.items():
            self.pool[k] = var

    def hydrate(self, deformat=False) -> Dict[str, Any]:
        """
        将当前上下文中的数据清洗成 Dict[str, Any] 类型的朴素数据，过程中会进行变量引用的分析和替换

        :param deformat: 是否将返回字典中的 key 值从 ${%s} 替换为 %s
        :type deformat: bool, optional
        :return: 上下文数据朴素值字典
        :rtype: Dict[str, Any]
        """
        key_formatter = deformat_var_key if deformat else _raw_key
        hydrated = {}

        for key, var in self.pool.items():
            hydrated[key_formatter(key)] = var.get()

        return hydrated

    def extract_outputs(
        self,
        pipeline_id: str,
        data_outputs: Dict[str, str],
        execution_data_outputs: Dict[str, Any],
    ):
        """
        将某个节点的输出提取到流程上下文中

        :param pipeline_id: 上下文对应的流程/子流程 ID
        :type pipeline_id: str
        :param data_outputs: 节点输出键映射
        :type data_outputs: Dict[str, str]
        :param execution_data_outputs: 节点执行数据输出
        :type execution_data_outputs: Dict[str, Any]
        """
        update = {}
        for origin_key, target_key in data_outputs.items():
            if origin_key not in execution_data_outputs:
                continue

            update[target_key] = ContextValue(
                key=target_key,
                type=ContextValueType.PLAIN,
                value=execution_data_outputs[origin_key],
            )

        self.runtime.upsert_plain_context_values(pipeline_id=pipeline_id, update=update)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from weakref import WeakValueDictionary
from datetime import datetime
from abc import ABCMeta, abstractmethod
from typing import List, Optional, Dict, Set, Any

from .models import (
    State,
    Node,
    Schedule,
    ScheduleType,
    Data,
    DataInput,
    ExecutionData,
    ExecutionHistory,
    ExecutionShortHistory,
    CallbackData,
    ProcessInfo,
    SuspendedProcessInfo,
    DispatchProcess,
    ContextValue,
)

# plugin interface


class Service(metaclass=ABCMeta):
    """
    服务对象接口
    """

    @abstractmethod
    def pre_execute(self, data: ExecutionData, root_pipeline_data: ExecutionData):
        """
        execute 执行前执行的逻辑

        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        """

    @abstractmethod
    def execute(self, data: ExecutionData, root_pipeline_data: ExecutionData) -> bool:
        """
        execute 逻辑

        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :return: 是否执行成功
        :rtype: bool
        """

    @abstractmethod
    def schedule(
        self,
        schedule: Schedule,
        data: ExecutionData,
        root_pipeline_data: ExecutionData,
        callback_data: Optional[CallbackData] = None,
    ) -> bool:
        """
        schedule 逻辑

        :param schedule: Schedule 对象
        :type schedule: Schedule
        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :param callback_data: 回调数据, defaults to None
        :type callback_data: Optional[CallbackData], optional
        :return: [description]
        :rtype: bool
        """

    @abstractmethod
    def need_schedule(self) -> bool:
        """
        服务是否需要调度

        :return: 是否需要调度
        :rtype: bool
        """

    @abstractmethod
    def schedule_type(self) -> Optional[ScheduleType]:
        """
        服务调度类型

        :return: 调度类型
        :rtype: Optional[ScheduleType]
        """

    @abstractmethod
    def is_schedule_done(self) -> bool:
        """
        调度是否完成

        :return: 调度是否完成
        :rtype: bool
        """

    @abstractmethod
    def schedule_after(
        self,
        schedule: Optional[Schedule],
        data: ExecutionData,
        root_pipeline_data: ExecutionData,
    ) -> int:
        """
        计算下一次调度间隔

        :param schedule: 调度对象，未进行调度时传入为空
        :type schedule: Optional[Schedule]
        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :return: 调度间隔，单位为秒
        :rtype: int
        """

    @abstractmethod
    def setup_runtime_attributes(self, **attrs):
        """
        装载运行时属性

        :param attrs: 运行时属性
        :type attrs: Dict[str, Any]
        """


class ExecutableEvent(metaclass=ABCMeta):
    """
    可执行结束节点接口
    """

    @abstractmethod
    def execute(pipeline_stack: List[str], root_pipeline_id: str):
        """
        execute 逻辑

        :param pipeline_stack: 流程栈
        :type pipeline_stack: List[str]
        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        """


class Variable(metaclass=ABCMeta):
    """
    变量接口
    """

    @abstractmethod
    def get(self) -> Any:
        """
        获取变量值

        :return: 变量值
        :rtype: Any
        """


# runtime interface


class PluginManagerMixin:
    """
    插件管理接口，声明了插件（服务，可执行结束节点，变量）管理相关的接口
    """

    @abstractmethod
    def get_service(self, code: str, version: str) -> Service:
        """
        根据代号与版本获取特定服务对象实例

        :param code: 服务唯一代号
        :type code: str
        :param version: 服务版本
        :type version: str
        :return: 服务对象实例
        :rtype: Service
        """

    @abstractmethod
    def get_executable_end_event(self, code: str) -> ExecutableEvent:
        """
        根据代号获取特定可执行结束事件实例

        :param code: 可执行结束事件唯一代号
        :type code: str
        :return: 可执行结束事件实例
        :rtype: ExecutableEvent:
        """

    @abstractmethod
    def get_compute_variable(
        self,
        code: str,
        key: str,
        value: Variable,
        additional_data: dict,
    ) -> Variable:
        """
        根据代号获取变量实例

        :param code: 唯一代号
        :type code: str
        :param key: 变量 key
        :type key: str
        :param value: 变量配置
        :type value: Any
        :param additional_data: 额外数据字典
        :type additional_data: dict
        :return: 变量实例
        :rtype: Variable
        """


class EngineAPIHooksMixin:
    """
    引擎 API 执行时调用的钩子相关接口声明
    """

    @abstractmethod
    def pre_prepare_run_pipeline(self, pipeline: dict, root_pipeline_data: dict, **options):
        """
        调用 pre_prepare_run_pipeline 前执行的钩子

        :param pipeline: 流程描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data dict
        """

    @abstractmethod
    def post_prepare_run_pipeline(self, pipeline: dict, root_pipeline_data: dict, **options):
        """
        调用 pre_prepare_run_pipeline 后执行的钩子

        :param pipeline: 流程描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data dict
        """

    @abstractmethod
    def pre_pause_pipeline(self, pipeline_id: str):
        """
        暂停 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def post_pause_pipeline(self, pipeline_id: str):
        """
        暂停 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def pre_revoke_pipeline(self, pipeline_id: str):
        """
        撤销 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def post_revoke_pipeline(self, pipeline_id: str):
        """
        撤销 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def pre_resume_pipeline(self, pipeline_id: str):
        """
        继续 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def post_resume_pipeline(self, pipeline_id: str):
        """
        继续 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    @abstractmethod
    def pre_resume_node(self, node_id: str):
        """
        继续节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def post_resume_node(self, node_id: str):
        """
        继续节点后执行的钩子

        :param node_id: [description]节点 ID
        :type node_id: str
        """

    @abstractmethod
    def pre_pause_node(self, node_id: str):
        """
        暂停节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def post_pause_node(self, node_id: str):
        """
        暂停节点后执行的钩子

        :param node_id: [description]节点 ID
        :type node_id: str
        """

    @abstractmethod
    def pre_retry_node(self, node_id: str, data: Optional[dict]):
        """
        重试节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 重试时使用的节点执行输入
        :type data: Optional[dict]
        """

    @abstractmethod
    def post_retry_node(self, node_id: str, data: Optional[dict]):
        """
        重试节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 重试时使用的节点执行输入
        :type data: Optional[dict]
        """

    @abstractmethod
    def pre_skip_node(self, node_id: str):
        """
        跳过节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def post_skip_node(self, node_id: str):
        """
        跳过节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def pre_skip_exclusive_gateway(self, node_id: str, flow_id: str):
        """
        跳过分支网关前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_id: 跳过后选择的目标流 ID
        :type flow_id: str
        """

    @abstractmethod
    def post_skip_exclusive_gateway(self, node_id: str, flow_id: str):
        """
        跳过分支网关后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_id: 跳过后选择的目标流 ID
        :type flow_id: str
        """

    @abstractmethod
    def pre_forced_fail_activity(self, node_id: str, ex_data: str):
        """
        强制失败节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param ex_data: 写入节点执行数据的失败信息
        :type ex_data: str
        """

    @abstractmethod
    def post_forced_fail_activity(self, node_id: str, ex_data: str, old_version: str, new_version: str):
        """
        强制失败节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param ex_data: 写入节点执行数据的失败信息
        :type ex_data: str
        :param old_version: 强制失败前的节点版本
        :type old_version: str
        :param new_version: 强制失败后的节点版本
        :type new_version: str
        """

    @abstractmethod
    def pre_callback(self, node_id: str, version: str, data: str):
        """
        回调节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: str
        """

    @abstractmethod
    def post_callback(self, node_id: str, version: str, data: str):
        """
        回调节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: str
        """


class TaskMixin:
    """
    引擎任务派发相关接口
    """

    @abstractmethod
    def execute(self, process_id: int, node_id: str):
        """
        派发执行任务，执行任务被拉起执行时应该调用 Engine 实例的 execute 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def schedule(
        self,
        process_id: int,
        node_id: str,
        schedule_id: str,
        callback_data_id: Optional[int] = None,
    ):
        """
        派发调度任务，调度任务被拉起执行时应该调用 Engine 实例的 schedule 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param schedule_id: 调度 ID
        :type schedule_id: str
        :param callback_data_id: 回调数据, defaults to None
        :type callback_data_id: Optional[int], optional
        """

    @abstractmethod
    def set_next_schedule(
        self,
        process_id: int,
        node_id: str,
        schedule_id: str,
        schedule_after: int,
        callback_data_id: Optional[int] = None,
    ):
        """
        设置下次调度时间，调度倒数归零后应该执行 Engine 实例的 schedule 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param schedule_id: 调度 ID
        :type schedule_id: str
        :param schedule_after: 调度倒数
        :type schedule_after: int
        :param callback_data_id: 回调数据, defaults to None
        :type callback_data_id: Optional[int], optional
        """

    @abstractmethod
    def start_timeout_monitor(self, process_id: int, node_id: str, version: str, timeout: int):
        """
        开始对某个节点执行的超时监控，若超时时间归零后节点未进入归档状态，则强制失败该节点

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        :param timeout: 超时时间，单位为秒
        :type timeout: int
        """

    @abstractmethod
    def stop_timeout_monitor(
        self,
        process_id: int,
        node_id: str,
        version: str,
    ):
        """
        停止对某个节点的超时监控

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        """


class ProcessMixin:
    """
    进程相关接口
    """

    @abstractmethod
    def beat(self, process_id: int):
        """
        进程心跳

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def wake_up(self, process_id: int):
        """
        将当前进程标记为唤醒状态

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def sleep(self, process_id: int):
        """
        将当前进程标记为睡眠状态

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def suspend(self, process_id: int, by: str):
        """
        将当前进程标记为阻塞状态

        :param process_id: 进程 ID
        :type process_id: int
        :param by: 造成阻塞的节点信息
        :type by: str
        """

    @abstractmethod
    def resume(self, process_id: int):
        """
        将进程标记为非阻塞状态

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def batch_resume(self, process_id_list: List[int]):
        """
        批量将进程标记为非阻塞状态

        :param process_id_list: 进程 ID 列表
        :type process_id_list: List[int]
        """

    @abstractmethod
    def die(self, process_id: int):
        """
        将当前进程标记为非存活状态

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def get_process_info(self, process_id: int) -> ProcessInfo:
        """
        获取某个进程的基本信息

        :param process_id: 进程 ID
        :type process_id: int
        :return: 进程基本信息
        :rtype: ProcessInfo
        """

    @abstractmethod
    def get_process_info_with_root_pipeline(self, pipeline_id: str) -> List[ProcessInfo]:
        """
        根据根流程 ID 获取一批进程的信息

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: 进程基本信息
        :rtype: List[ProcessInfo]
        """

    @abstractmethod
    def kill(self, process_id: int):
        """
        强制结束某个进程正在进行的活动，并将其标志为睡眠状态

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def get_suspended_process_info(self, suspended_by: str) -> List[SuspendedProcessInfo]:
        """
        获取由于 pipeline 暂停而被暂停执行的进程信息

        : param suspended_by: 进程 ID
        : type suspended_by: str
        : return: 暂停的进程信息
        : rtype: SuspendedProcessInfo
        """

    @abstractmethod
    def get_sleep_process_with_current_node_id(self, node_id: str) -> Optional[int]:
        """
        获取由于处于睡眠状态且当前节点 ID 为 node_id 的进程 ID

        : param node_id: 节点 ID
        : type node_id: str
        : return: 进程 ID
        : rtype: str
        """

    @abstractmethod
    def get_process_id_with_current_node_id(self, node_id: str) -> Optional[int]:
        """
        获取当前节点 ID 为 node_id 且存活的进程 ID

        : param node_id: 节点 ID
        : type node_id: str
        : return: 进程 ID
        : rtype: str
        """

    @abstractmethod
    def set_current_node(self, process_id: int, node_id: str):
        """
        将进程当前处理节点标记为 node

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        """

    @abstractmethod
    def child_process_finish(self, parent_id: int, process_id: int) -> bool:
        """
        标记某个进程的子进程执行完成，并返回是否能够唤醒父进程继续执行的标志位

        :param parent_id: 父进程 ID
        :type parent_id: int
        :param process_id: 子进程 ID
        :type process_id: int
        :return: 是否能够唤醒父进程继续执行
        :rtype: bool
        """

    @abstractmethod
    def is_frozen(self, process_id: int) -> bool:
        """
        检测当前进程是否需要被冻结

        :param process_id: 进程 ID
        :type process_id: int
        :return: 是否需要被冻结
        :rtype: bool
        """

    @abstractmethod
    def freeze(self, process_id: int):
        """
        冻结当前进程

        :param process_id: 进程 ID
        :type process_id: int
        """

    @abstractmethod
    def fork(
        self,
        parent_id: str,
        root_pipeline_id: str,
        pipeline_stack: List[str],
        from_to: Dict[str, str],
    ) -> List[DispatchProcess]:
        """
        根据当前进程 fork 出多个子进程

        :param parent_id: 父进程 ID
        :type parent_id: str
        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param pipeline_stack: 子流程栈
        :type pipeline_stack: List[str]
        :param from_to: 子进程的执行开始节点和目标节点
        :type from_to: Dict[str, str]
        :return: 待调度进程信息列表
        :rtype: List[DispatchProcess]
        """

    @abstractmethod
    def join(self, process_id: int, children_id: List[str]):
        """
        让父进程等待子进程

        :param process_id: 父进程 ID
        :type process_id: int
        :param children_id: 子进程 ID 列表
        :type children_id: List[str]
        """

    @abstractmethod
    def set_pipeline_stack(self, process_id: int, stack: List[str]):
        """
        设置进程的流程栈

        :param process_id: 进程 ID
        :type process_id: int
        :param stack: 流程栈
        :type stack: List[str]
        """


class StateMixin:
    """
    状态相关接口
    """

    @abstractmethod
    def get_state(self, node_id: str) -> State:
        """
        获取某个节点的状态对象

        : param node_id: 节点 ID
        : type node_id: str
        : return: State 实例
        : rtype: State
        """

    @abstractmethod
    def get_state_or_none(self, node_id: str) -> Optional[State]:
        """
        获取某个节点的状态对象，如果不存在则返回 None

        : param node_id: 节点 ID
        : type node_id: str
        : return: State 实例
        : rtype: State
        """

    @abstractmethod
    def get_state_by_root(self, root_id: str) -> List[State]:
        """
        根据根节点 ID 获取一批节点状态

        :param root_id: 根节点 ID
        :type root_id: str
        :return: 节点状态列表
        :rtype: List[State]
        """

    @abstractmethod
    def get_state_by_parent(self, parent_id: str) -> List[State]:
        """
        根据父节点 ID 获取一批节点状态

        :param parent_id: 父节点 ID
        :type parent_id: str
        :return: 节点状态列表
        :rtype: List[State]
        """

    @abstractmethod
    def batch_get_state_name(self, node_id_list: List[str]) -> Dict[str, str]:
        """
        批量获取一批节点的状态

        :param node_id_list: 节点 ID 列表
        :type node_id_list: List[str]
        :return: 节点ID -> 状态名称
        :rtype: Dict[str, str]
        """

    @abstractmethod
    def has_state(self, node_id: str) -> bool:
        """
        是否存在某个节点的的状态

        :param node_id: 节点 ID
        :type node_id: str
        :return: 该节点状态是否存在
        :rtype: bool
        """

    @abstractmethod
    def set_state(
        self,
        node_id: str,
        to_state: str,
        loop: int = -1,
        version: str = None,
        root_id: Optional[str] = None,
        parent_id: Optional[str] = None,
        is_retry: bool = False,
        is_skip: bool = False,
        reset_retry: bool = False,
        reset_skip: bool = False,
        error_ignored: bool = False,
        reset_error_ignored: bool = False,
        refresh_version: bool = False,
        clear_started_time: bool = False,
        set_started_time: bool = False,
        clear_archived_time: bool = False,
        set_archive_time: bool = False,
    ) -> str:
        """
        设置节点的状态，如果节点存在，进行状态转换时需要满足状态转换状态机

        :param node_id: 节点 ID
        :type node_id: str
        :param to_state: 目标状态
        :type to_state: str
        :param loop: 循环次数, 为 -1 时表示不设置
        :type loop: int, optional
        :param version: 目标状态版本，为空时表示不做版本校验
        :type version: Optional[str], optional
        :param root_id: 根节点 ID，为空时表示不设置
        :type root_id: Optional[str], optional
        :param parent_id: 父节点 ID，为空时表示不设置
        :type parent_id: Optional[str], optional
        :param is_retry: 是否增加重试次数
        :type is_retry: bool, optional
        :param is_skip: 是否将跳过设置为 True
        :type is_skip: bool, optional
        :param reset_retry: 是否重置重试次数
        :type reset_retry: bool, optional
        :param reset_skip: 是否重置跳过标志
        :type reset_skip: bool, optional
        :param error_ignored: 是否为忽略错误跳过
        :type error_ignored: bool, optional
        :param reset_error_ignored: 是否重置忽略错误标志
        :type reset_error_ignored: bool, optional
        :param refresh_version: 是否刷新版本号
        :type refresh_version: bool, optional
        :param clear_started_time: 是否清空开始时间
        :type clear_started_time: bool, optional
        :param set_started_time: 是否设置开始时间
        :type set_started_time: bool, optional
        :param clear_archived_time: 是否清空归档时间
        :type clear_archived_time: bool, optional
        :param set_archive_time: 是否设置归档时间
        :type set_archive_time: bool, optional
        :return: 该节点最新版本
        :rtype: str
        """

    @abstractmethod
    def set_state_root_and_parent(self, node_id: str, root_id: str, parent_id: str):
        """
        设置节点的根流程和父流程 ID

        :param node_id: 节点 ID
        :type node_id: str
        :param root_id: 根流程 ID
        :type root_id: str
        :param parent_id: 父流程 ID
        :type parent_id: str
        """


class NodeMixin:
    """
    节点相关接口
    """

    @abstractmethod
    def get_node(self, node_id: str) -> Node:
        """
        获取某个节点的详细信息

        :param node_id: 节点 ID
        :type node_id: str
        :return: Node 实例
        :rtype: Node
        """


class ScheduleMixin:
    """
    调度实例相关接口
    """

    @abstractmethod
    def set_schedule(
        self,
        process_id: int,
        node_id: str,
        version: str,
        schedule_type: ScheduleType,
    ) -> Schedule:
        """
        设置 schedule 对象

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        :param schedule_type: 调度类型
        :type schedule_type: ScheduleType
        :return: 调度对象实例
        :rtype: Schedule
        """

    @abstractmethod
    def get_schedule(self, schedule_id: str) -> Schedule:
        """
        获取 Schedule 对象

        :param schedule_id: 调度实例 ID
        :type schedule_id: str
        :return: Schedule 对象实例
        :rtype: Schedule
        """

    @abstractmethod
    def get_schedule_with_node_and_version(self, node_id: str, version: str) -> Schedule:
        """
        通过节点 ID 和执行版本来获取 Scheudle 对象

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        :return: Schedule 对象
        :rtype: Schedule
        """

    @abstractmethod
    def apply_schedule_lock(self, schedule_id: str) -> bool:
        """
        获取 Schedule 对象的调度锁，返回是否成功获取锁

        :param schedule_id: 调度实例 ID
        :type schedule_id: str
        :return: 是否成功获取锁
        :rtype: bool
        """

    @abstractmethod
    def release_schedule_lock(self, schedule_id: int):
        """
        释放指定 Schedule 的调度锁

        :param schedule_id: Schedule ID
        :type schedule_id: int
        """

    @abstractmethod
    def expire_schedule(self, schedule_id: int):
        """
        将某个 Schedule 对象标记为已过期

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """

    @abstractmethod
    def finish_schedule(self, schedule_id: int):
        """
        将某个 Schedule 对象标记为已完成

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """

    @abstractmethod
    def add_schedule_times(self, schedule_id: int):
        """
        将某个 Schedule 对象的调度次数 +1

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """


class ContextMixin:
    """
    流程上下文相关接口
    """

    @abstractmethod
    def get_context_values(self, pipeline_id: str, keys: set) -> List[ContextValue]:
        """
        获取某个流程上下文中的 keys 所指定的键对应变量的值

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param keys: 变量键
        :type keys: set
        :return: 变量值信息
        :rtype: List[ContextValue]
        """

    @abstractmethod
    def get_context_key_references(self, pipeline_id: str, keys: set) -> set:
        """
        获取某个流程上下文中 keys 所指定的变量直接和间接引用的其他所有变量的键

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param keys: 变量 key 列表
        :type keys: set
        :return: keys 所指定的变量直接和简介引用的其他所有变量的键
        :rtype: set
        """

    @abstractmethod
    def upsert_plain_context_values(self, pipeline_id: str, update: Dict[str, ContextValue]):
        """
        更新或创建新的普通上下文数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param update: 更新数据
        :type update: Dict[str, ContextValue]
        """

    def get_context(self, pipeline_id: str) -> List[ContextValue]:
        """
        获取某个流程的所有上下文数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: [description]
        :rtype: List[ContextValue]
        """

    def get_context_outputs(self, pipeline_id: str) -> Set[str]:
        """
        获取流程上下文需要输出的数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: 输出数据 key
        :rtype: Set[str]
        """


class DataMixin:
    """
    节点数据，执行数据，回调数据相关接口
    """

    @abstractmethod
    def get_data(self, node_id: str) -> Data:
        """
        获取某个节点的数据对象

        :param node_id: 节点 ID
        :type node_id: str
        :return: 数据对象实例
        :rtype: Data
        """

    @abstractmethod
    def get_data_inputs(self, node_id: str) -> Dict[str, DataInput]:
        """
        获取某个节点的输入数据

        :param node_id: 节点 ID
        :type node_id: str
        :return: 输入数据字典
        :rtype: dict
        """

    @abstractmethod
    def get_data_outputs(self, node_id: str) -> dict:
        """
        获取某个节点的输出数据

        :param node_id: 节点 ID
        :type node_id: str
        :return: 输入数据字典
        :rtype: dict
        """

    @abstractmethod
    def set_data_inputs(self, node_id: str, data: Dict[str, DataInput]):
        """
        将节点数据对象的 inputs 设置为 data

        : param node_id: 节点 ID
        : type node_id: str
        : param data: 目标数据
        : type data: dict
        """

    # execution data relate
    @abstractmethod
    def get_execution_data(self, node_id: str) -> ExecutionData:
        """
        获取某个节点的执行数据

        : param node_id: 节点 ID
        : type node_id: str
        : return: 执行数据实例
        : rtype: ExecutionData
        """

    @abstractmethod
    def get_execution_data_inputs(self, node_id: str) -> dict:
        """
        获取某个节点的执行数据输入

        :param node_id: 节点 ID
        :type node_id: str
        :return: 执行数据输入
        :rtype: dict
        """

    @abstractmethod
    def get_execution_data_outputs(self, node_id: str) -> dict:
        """
        获取某个节点的执行数据输出

        :param node_id: 节点 ID
        :type node_id: str
        :return: 执行数据输出
        :rtype: dict
        """

    @abstractmethod
    def set_execution_data(self, node_id: str, data: ExecutionData):
        """
        设置某个节点的执行数据

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 执行数据实例
        :type data: ExecutionData
        """

    @abstractmethod
    def set_execution_data_inputs(self, node_id: str, inputs: dict):
        """
        设置某个节点的执行数据输入

        :param node_id: 节点 ID
        :type node_id: str
        :param outputs: 输出数据
        :type outputs: dict
        """

    @abstractmethod
    def set_execution_data_outputs(self, node_id: str, outputs: dict):
        """
        设置某个节点的执行数据输出

        :param node_id: 节点 ID
        :type node_id: str
        :param outputs: 输出数据
        :type outputs: dict
        """

    # callback data relate
    @abstractmethod
    def set_callback_data(self, node_id: str, version: str, data: dict) -> int:
        """
        设置某个节点执行数据的回调数据

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: dict
        :return: 回调数据 ID
        :rtype: int
        """

    @abstractmethod
    def get_callback_data(self, data_id: int) -> CallbackData:
        """
        获取回调数据

        :param data_id: Data ID
        :type data_id: int
        :return: 回调数据实例
        :rtype: CallbackData
        """


class ExecutionHistoryMixin:
    """
    执行历史相关接口
    """

    @abstractmethod
    def add_history(
        self,
        node_id: str,
        started_time: datetime,
        archived_time: datetime,
        loop: int,
        skip: bool,
        retry: int,
        version: str,
        inputs: dict,
        outputs: dict,
    ) -> int:
        """
        为某个节点记录一次执行历史

        : param node_id: 节点 ID
        : type node_id: str
        : param started_time: 开始时间
        : type started_time: datetime
        : param archived_time: 归档时间
        : type archived_time: datetime
        : param loop: 重入计数
        : type loop: int
        : param skip: 是否跳过
        : type skip: bool
        : param retry: 重试次数
        : type retry: int
        : param version: 节点执行版本号
        : type version: str
        : param inputs: 输入数据
        : type inputs: dict
        : param outputs: 输出数据
        : type outputs: dict
        """

    @abstractmethod
    def get_histories(self, node_id: str, loop: int = -1) -> List[ExecutionHistory]:
        """
        返回某个节点的历史记录

        :param node_id: 节点 ID
        :type node_id: str
        :param loop: 重入次数, -1 表示不过滤重入次数
        :type loop: int, optional
        :return: 历史记录列表
        :rtype: List[History]
        """

    @abstractmethod
    def get_short_histories(self, node_id: str, loop: int = -1) -> List[ExecutionShortHistory]:
        """
        返回某个节点的简要历史记录

        :param node_id: 节点 ID
        :type node_id: str
        :param loop: 重入次数, -1 表示不过滤重入次数
        :type loop: int, optional
        :return: 历史记录列表
        :rtype: List[ExecutionShortHistory]
        """


class EngineRuntimeInterface(
    PluginManagerMixin,
    EngineAPIHooksMixin,
    TaskMixin,
    ProcessMixin,
    StateMixin,
    NodeMixin,
    ScheduleMixin,
    ContextMixin,
    DataMixin,
    ExecutionHistoryMixin,
    metaclass=ABCMeta,
):
    @abstractmethod
    def prepare_run_pipeline(self, pipeline: dict, root_pipeline_data: dict, **options) -> int:
        """
        进行 pipeline 执行前的准备工作，并返回 进程 ID，该函数执行完成后即代表
        pipeline 是随时可以通过 execute(process_id, start_event_id) 启动执行的
        一般来说，应该完成以下工作:
        - 准备好进程模型
        - 准备好流程中每个节点的信息
        - 准备好流程中每个节点数据对象的信息

        :param pipeline: pipeline 描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data dict
        :return: 进程 ID
        :rtype: str
        """

    @abstractmethod
    def node_rerun_limit(self, root_pipeline_id: str, node_id: str) -> int:
        """
        返回节点最大重入次数

        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param node_id: 节点 ID
        :type node_id: str
        :return: 节点最大重入次数
        :rtype: int
        """



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
节点处理器逻辑封装模块
"""

from typing import Optional, List
from abc import ABCMeta, abstractmethod

from bamboo_engine import states

from .eri import (
    EngineRuntimeInterface,
    Node,
    Schedule,
    CallbackData,
    ScheduleType,
    DispatchProcess,
    ProcessInfo,
    NodeType,
)
from .exceptions import NotFoundError


def register_handler(type: NodeType):
    """
    节点 Handler 注册函数

    :param type: 节点类型
    :type type: NodeType
    """

    def register(cls):
        HandlerFactory.add_handler(type, cls)
        return cls

    return register


class ExecuteResult:
    """
    Handler execute 方法返回的结果
    """

    def __init__(
        self,
        should_sleep: bool,
        schedule_ready: bool,
        schedule_type: Optional[ScheduleType],
        schedule_after: int,
        dispatch_processes: List[DispatchProcess],
        next_node_id: Optional[str],
        should_die: bool = False,
    ):
        """

        :param should_sleep: 当前进程是否应该进入休眠
        :type should_sleep: bool
        :param schedule_ready: 被处理的节点是否准备好进入调度
        :type schedule_ready: bool
        :param schedule_type: 被处理的节点的调度类型
        :type schedule_type: Optional[ScheduleType]
        :param schedule_after: 在 schedule_after 秒后开始执行调度
        :type schedule_after: int
        :param dispatch_processes: 需要派发的子进程信息列表
        :type dispatch_processes: List[DispatchProcess]
        :param next_node_id: 推进循环中下一个要处理的节点的 ID
        :type next_node_id: Optional[str]
        :param should_die: 当前进程是否需要进入死亡状态, defaults to False
        :type should_die: bool, optional
        """
        self.should_sleep = should_sleep
        self.schedule_ready = schedule_ready
        self.schedule_type = schedule_type
        self.schedule_after = schedule_after
        self.dispatch_processes = dispatch_processes
        self.next_node_id = next_node_id
        self.should_die = should_die


class ScheduleResult:
    """
    Handler schedule 方法返回的结果
    """

    def __init__(
        self,
        has_next_schedule: bool,
        schedule_after: int,
        schedule_done: bool,
        next_node_id: Optional[str],
    ):
        """

        :param has_next_schedule: 是否还有下次调度
        :type has_next_schedule: bool
        :param schedule_after: 在 schedule_after 秒后开始下次调度
        :type schedule_after: int
        :param schedule_done: 调度是否完成
        :type schedule_done: bool
        :param next_node_id: 调度完成后下一个需要执行的节点的 ID
        :type next_node_id: Optional[str]
        """
        self.has_next_schedule = has_next_schedule
        self.schedule_after = schedule_after
        self.schedule_done = schedule_done
        self.next_node_id = next_node_id


class NodeHandler(metaclass=ABCMeta):
    """
    节点处理器，负责封装不同类型节点的 execute 和 schedule 逻辑
    """

    LOOP_KEY = "_loop"

    def __init__(self, node: Node, runtime: EngineRuntimeInterface):
        """

        :param node: 节点实例
        :type node: Node
        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        """
        self.node = node
        self.runtime = runtime

    @abstractmethod
    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :param loop: 重入次数
        :type loop: int
        :param version: 执行版本
        :type version: str
        :return: 执行结果
        :rtype: ExecuteResult
        """

    def schedule(
        self,
        process_info: ProcessInfo,
        loop: int,
        schedule: Schedule,
        callback_data: Optional[CallbackData] = None,
    ) -> ScheduleResult:
        """
        节点的 schedule 处理逻辑，不支持 schedule 的节点可以不实现该方法

        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :param loop: 重入次数
        :type loop: int
        :param schedule: Schedule 实例
        :type schedule: Schedule
        :param callback_data: 回调数据, defaults to None
        :type callback_data: Optional[CallbackData], optional
        :return: 调度结果
        :rtype: ScheduleResult
        """
        raise NotImplementedError()

    def _execute_fail(self, ex_data: str) -> ExecuteResult:
        exec_outputs = self.runtime.get_execution_data_outputs(self.node.id)

        self.runtime.set_state(
            node_id=self.node.id, to_state=states.FAILED, set_archive_time=True
        )

        exec_outputs["ex_data"] = ex_data

        self.runtime.set_execution_data_outputs(self.node.id, exec_outputs)

        return ExecuteResult(
            should_sleep=True,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=[],
            next_node_id=None,
        )

    def _get_plain_inputs(self, node_id: str):
        return {
            key: di.value for key, di in self.runtime.get_data_inputs(node_id).items()
        }


class HandlerFactory:
    """
    节点处理器工厂
    """

    _handlers = {}

    @classmethod
    def add_handler(cls, type: NodeType, handler_cls):
        """
        向工厂中注册某个类型节点的处理器

        :param type: 节点类型
        :type type: NodeType
        :param handler_cls: [description]
        :type handler_cls: [type]
        :raises InvalidOperationError: [description]
        """
        if not issubclass(handler_cls, NodeHandler):
            raise InvalidOperationError(
                "register handler err: {} is not subclass of {}".format(
                    handler_cls, "NodeHandler"
                )
            )
        cls._handlers[type.value] = handler_cls

    @classmethod
    def get_handler(cls, node: Node, runtime: EngineRuntimeInterface) -> NodeHandler:
        """
        获取某个节点的处理器实例

        :param node: 节点实例
        :type node: NodeType
        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :raises NotFoundError: [description]
        :return: 节点处理器实例
        :rtype: NodeHandler
        """
        if node.type.value not in cls._handlers:
            raise NotFoundError(
                "can not find handler for {} type node".format(node.type.value)
            )

        return cls._handlers[node.type.value](node, runtime)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))
about = {}
with open(path.join(here, "bamboo_engine", "__version__.py"), "r") as f:
    exec(f.read(), about)

long_description = """
Bamboo-engine is a general-purpose process engine which can parse, 
execute and schedule process tasks created by users, and provides flexible control capabilities such as pause, revoke, skip, force failure, retry and re-execute,
and advanced features such as parallelism and sub-processes. 
It can further improve the concurrent processing ability of tasks through horizontal expansion.
"""
version = about["__version__"]

setup(
    name="bamboo-engine",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,
    description="bamboo-engine",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/Tencent/bk-sops/tree/sdk/sdk/bamboo-engine",
    # Author details
    author="Blueking",
    author_email="howellliang@tencent.com",
    include_package_data=True,
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "Werkzeug>=1.0.1,<2.0",
        "pyparsing>=2.2.0,<3.0",
        "mako>=1.1.4,<2.0",
        "prometheus-client>=0.9.0,<1.0.0",
    ],
    zip_safe=False,
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     "console_scripts": ["bk-admin=blueapps.contrib.bk_commands:bk_admin"],
    # },
)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
API 模块用于向外暴露接口，bamboo-engine 的使用者应该永远只用这个模块与 bamboo-engien 进行交互
"""

import logging
import functools
import traceback
from typing import Optional, Any
from contextlib import contextmanager

from .utils.object import Representable
from .eri import EngineRuntimeInterface
from .engine import Engine

logger = logging.getLogger("bamboo_engine")


class EngineAPIResult(Representable):
    """
    api 统一返回结果
    """

    def __init__(
        self,
        result: bool,
        message: str,
        exc: Optional[Exception] = None,
        data: Optional[Any] = None,
        exc_trace: Optional[str] = None,
    ):
        """
        :param result: 是否执行成功
        :type result: bool
        :param message: 附加消息，result 为 False 时关注
        :type message: str
        :param exc: 异常对象
        :type exc: Exception
        :param data: 数据
        :type data: Any
        """
        self.result = result
        self.message = message
        self.exc = exc
        self.data = data
        self.exc_trace = exc_trace


def _ensure_return_api_result(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            data = func(*args, **kwargs)
        except Exception as e:
            logger.exception("{} raise error.".format(func.__name__))
            trace = traceback.format_exc()
            return EngineAPIResult(result=False, message="fail", exc=e, data=None, exc_trace=trace)

        return EngineAPIResult(result=True, message="success", exc=None, data=data, exc_trace=None)

    return wrapper


@_ensure_return_api_result
def run_pipeline(runtime: EngineRuntimeInterface, pipeline: dict, **options) -> EngineAPIResult:
    """
    执行 pipeline

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param pipeline: pipeline 描述对象
    :type pipeline: dict
    :return: 执行结果
    :rtype: EngineAPIResult
    """

    Engine(runtime).run_pipeline(pipeline, **options)


@_ensure_return_api_result
def pause_pipeline(runtime: EngineRuntimeInterface, pipeline_id: str) -> EngineAPIResult:
    """
    暂停 pipeline 的执行

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param pipeline_id: piipeline id
    :type pipeline_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """

    Engine(runtime).pause_pipeline(pipeline_id)


@_ensure_return_api_result
def revoke_pipeline(runtime: EngineRuntimeInterface, pipeline_id: str) -> EngineAPIResult:
    """
    撤销 pipeline，使其无法继续执行

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param pipeline_id: pipeline id
    :type pipeline_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).revoke_pipeline(pipeline_id)


@_ensure_return_api_result
def resume_pipeline(runtime: EngineRuntimeInterface, pipeline_id: str) -> EngineAPIResult:
    """
    继续被 pause_pipeline 接口暂停的 pipeline 的执行

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param pipeline_id: pipeline id
    :type pipeline_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).resume_pipeline(pipeline_id)


@_ensure_return_api_result
def pause_node_appoint(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    预约暂停某个节点的执行

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 id
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).pause_node_appoint(node_id)


@_ensure_return_api_result
def resume_node_appoint(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    继续由于某个节点而暂停的 pipeline 的执行

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 id
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).resume_node_appoint(node_id)


@_ensure_return_api_result
def retry_node(runtime: EngineRuntimeInterface, node_id: str, data: Optional[dict] = None) -> EngineAPIResult:
    """
    重试某个执行失败的节点

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 失败的节点 id
    :type node_id: str
    :param data: 重试时使用的节点执行输入
    :type data: dict
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).retry_node(node_id, data)


@_ensure_return_api_result
def skip_node(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    跳过某个执行失败的节点（仅限 event，activity）

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 失败的节点 id
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).skip_node(node_id)


@_ensure_return_api_result
def skip_exclusive_gateway(runtime: EngineRuntimeInterface, node_id: str, flow_id: str) -> EngineAPIResult:
    """
    跳过某个执行失败的分支网关

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 失败的分支网关 id
    :type node_id: str
    :param flow_id: 需要往下执行的 flow id
    :type flow_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).skip_exclusive_gateway(node_id, flow_id)


@_ensure_return_api_result
def forced_fail_activity(runtime: EngineRuntimeInterface, node_id: str, ex_data: str) -> EngineAPIResult:
    """
    强制失败某个 activity 节点

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :param message: 异常信息
    :type message: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).forced_fail_activity(node_id, ex_data)


@_ensure_return_api_result
def callback(runtime: EngineRuntimeInterface, node_id: str, version: str, data: dict) -> EngineAPIResult:
    """
    回调某个节点

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param version: 节点执行版本
    :param version: str
    :param data: 节点 ID
    :type data: dict
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    Engine(runtime).callback(node_id, version, data)


@_ensure_return_api_result
def get_pipeline_states(runtime: EngineRuntimeInterface, root_id: str, flat_children=True) -> EngineAPIResult:
    """
    返回某个任务的状态树

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param root_id: 根节点 ID
    :type root_id: str
    :param flat_children: 是否将所有子节点展开
    :type flat_children: bool
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    states = runtime.get_state_by_root(root_id)
    if not states:
        return {}

    root_state = None
    children = {}
    for s in states:
        if s.node_id != root_id:
            children[s.node_id] = {
                "id": s.node_id,
                "state": s.name,
                "root_id:": s.root_id,
                "parent_id": s.parent_id,
                "version": s.version,
                "loop": s.loop,
                "retry": s.retry,
                "skip": s.skip,
                "error_ignorable": s.error_ignored,
                "error_ignored": s.error_ignored,
                "created_time": s.created_time,
                "started_time": s.started_time,
                "archived_time": s.archived_time,
                "children": {},
            }
        else:
            root_state = s

    if not flat_children:
        # set node children
        for node_id, state in children.items():
            if state["parent_id"] in children:
                children[state["parent_id"]]["children"][node_id] = state

        # pop sub child
        for node_id in list(children.keys()):
            if children[node_id]["parent_id"] != root_state.node_id:
                children.pop(node_id)

    state_tree = {}
    state_tree[root_state.node_id] = {
        "id": root_state.node_id,
        "state": root_state.name,
        "root_id:": root_state.root_id,
        "parent_id": root_state.root_id,
        "version": root_state.version,
        "loop": root_state.loop,
        "retry": root_state.retry,
        "skip": root_state.skip,
        "error_ignorable": s.error_ignored,
        "error_ignored": s.error_ignored,
        "created_time": root_state.created_time,
        "started_time": root_state.started_time,
        "archived_time": root_state.archived_time,
        "children": children,
    }
    return state_tree


@_ensure_return_api_result
def get_children_states(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    返回某个节点及其所有子节点的状态

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 父流程 ID
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    parent_state = runtime.get_state_or_none(node_id)
    if not parent_state:
        return {}

    states = runtime.get_state_by_parent(node_id)
    children = {}
    for s in states:
        children[s.node_id] = {
            "id": s.node_id,
            "state": s.name,
            "root_id:": s.root_id,
            "parent_id": s.parent_id,
            "version": s.version,
            "loop": s.loop,
            "retry": s.retry,
            "skip": s.skip,
            "error_ignorable": s.error_ignored,
            "error_ignored": s.error_ignored,
            "created_time": s.created_time,
            "started_time": s.started_time,
            "archived_time": s.archived_time,
            "children": {},
        }

    state_tree = {}
    state_tree[parent_state.node_id] = {
        "id": parent_state.node_id,
        "state": parent_state.name,
        "root_id:": parent_state.root_id,
        "parent_id": parent_state.root_id,
        "version": parent_state.version,
        "loop": parent_state.loop,
        "retry": parent_state.retry,
        "skip": parent_state.skip,
        "error_ignorable": parent_state.error_ignored,
        "error_ignored": parent_state.error_ignored,
        "created_time": parent_state.created_time,
        "started_time": parent_state.started_time,
        "archived_time": parent_state.archived_time,
        "children": children,
    }
    return state_tree


@_ensure_return_api_result
def get_execution_data_inputs(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    获取某个节点执行数据的输入数据

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    return runtime.get_execution_data_inputs(node_id)


@_ensure_return_api_result
def get_execution_data_outputs(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    获取某个节点的执行数据输出

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    return runtime.get_execution_data_outputs(node_id)


@_ensure_return_api_result
def get_execution_data(runtime: EngineRuntimeInterface, node_id: str) -> EngineAPIResult:
    """
    获取某个节点的执行数据

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    data = runtime.get_execution_data(node_id)
    return {"inputs": data.inputs, "outputs": data.outputs}


@_ensure_return_api_result
def get_node_histories(runtime: EngineRuntimeInterface, node_id: str, loop: int = -1) -> EngineAPIResult:
    """
    获取某个节点的历史记录概览

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :param loop: 重入次数, -1 表示不过滤重入次数
    :type loop: int, optional
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    return [
        {
            "id": h.id,
            "node_id": h.node_id,
            "started_time": h.started_time,
            "archived_time": h.archived_time,
            "loop": h.loop,
            "skip": h.skip,
            "version": h.version,
            "inputs": h.inputs,
            "outputs": h.outputs,
        }
        for h in runtime.get_histories(node_id, loop)
    ]


@_ensure_return_api_result
def get_node_short_histories(runtime: EngineRuntimeInterface, node_id: str, loop: int = -1) -> EngineAPIResult:
    """
    获取某个节点的简要历史记录

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :param loop: 重入次数, -1 表示不过滤重入次数
    :type loop: int, optional
    :return: 执行结果
    :rtype: EngineAPIResult
    """
    return [
        {
            "id": h.id,
            "node_id": h.node_id,
            "started_time": h.started_time,
            "archived_time": h.archived_time,
            "loop": h.loop,
            "skip": h.skip,
            "version": h.version,
        }
        for h in runtime.get_short_histories(node_id, loop)
    ]


@_ensure_return_api_result
def get_pipeline_debug_info(self, runtime: EngineRuntimeInterface, pipeline_id: str):
    """
    获取某个流程的调试信息

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param pipeline_id: 流程 ID
    :type pipeline_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """

    return {
        "contex_values": runtime.get_context(pipeline_id),
        "processes": runtime.get_process_info_with_root_pipeline(pipeline_id),
    }


@_ensure_return_api_result
def get_node_debug_info(self, runtime: EngineRuntimeInterface, node_id: str):
    """
    获取某个节点的调试信息

    :param runtime: 引擎运行时实例
    :type runtime: EngineRuntimeInterface
    :param node_id: 节点 ID
    :type node_id: str
    :return: 执行结果
    :rtype: EngineAPIResult
    """

    data = None
    state = None
    err = []

    try:
        data = runtime.get_data(node_id)
    except Exception as e:
        err.append(str(e))

    try:
        state = runtime.get_state(node_id)
    except Exception as e:
        err.append(str(e))

    return {
        "node": runtime.get_node(node_id),
        "data": data,
        "state": state,
        "err": err,
    }



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
引擎核心模块
"""

import time
import random
import logging
import traceback
from typing import Optional
from contextlib import contextmanager


from . import states
from . import validator
from .local import set_node_info, CurrentNodeInfo
from .exceptions import InvalidOperationError, NotFoundError, StateVersionNotMatchError
from .handler import HandlerFactory
from .metrics import (
    ENGINE_RUNNING_PROCESSES,
    ENGINE_RUNNING_SCHEDULES,
    ENGINE_PROCESS_RUNNING_TIME,
    ENGINE_SCHEDULE_RUNNING_TIME,
    ENGINE_NODE_EXECUTE_TIME,
    ENGINE_NODE_SCHEDULE_TIME,
    setup_gauge,
    setup_histogram,
)
from .eri import (
    EngineRuntimeInterface,
    ScheduleType,
    NodeType,
    NodeType,
    ProcessInfo,
    State,
    ExecutionData,
    DataInput,
    Node,
)
from .utils.string import get_lower_case_name

logger = logging.getLogger("bamboo_engine")


class Engine:
    """
    流程引擎，封装流程核心调度逻辑
    """

    PURE_SKIP_ENABLE_NODE_TYPE = {NodeType.ServiceActivity, NodeType.EmptyStartEvent}

    def __init__(self, runtime: EngineRuntimeInterface):
        self.runtime = runtime

    # api
    def run_pipeline(self, pipeline: dict, root_pipeline_data: Optional[dict] = None, **options):
        """
        运行流程

        :param pipeline: 流程数据
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data dict
        """

        root_pipeline_data = {} if root_pipeline_data is None else root_pipeline_data
        cycle_tolerate = options.get("cycle_tolerate", False)
        validator.validate_and_process_pipeline(pipeline, cycle_tolerate)

        self.runtime.pre_prepare_run_pipeline(pipeline, root_pipeline_data, **options)

        process_id = self.runtime.prepare_run_pipeline(pipeline, root_pipeline_data, **options)
        # execute from start event
        self.runtime.execute(process_id, pipeline["start_event"]["id"])

        self.runtime.post_prepare_run_pipeline(pipeline, root_pipeline_data, **options)

    def pause_pipeline(self, pipeline_id: str):
        """
        暂停流程

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """
        if not self.runtime.has_state(pipeline_id):
            raise NotFoundError("node({}) does not exist".format(pipeline_id))

        self.runtime.pre_pause_pipeline(pipeline_id)

        self.runtime.set_state(node_id=pipeline_id, to_state=states.SUSPENDED)

        self.runtime.post_pause_pipeline(pipeline_id)

    def revoke_pipeline(self, pipeline_id: str):
        """
        撤销流程

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """
        if not self.runtime.has_state(pipeline_id):
            raise NotFoundError("node({}) does not exist".format(pipeline_id))

        self.runtime.pre_revoke_pipeline(pipeline_id)

        self.runtime.set_state(node_id=pipeline_id, to_state=states.REVOKED)

        self.runtime.post_revoke_pipeline(pipeline_id)

    def resume_pipeline(self, pipeline_id: str):
        """
        继续流程

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """
        state = self.runtime.get_state(pipeline_id)

        if state.name != states.SUSPENDED:
            raise InvalidOperationError("pipeline({}) state is: {}".format(pipeline_id, state.name))

        info_list = self.runtime.get_suspended_process_info(pipeline_id)

        self.runtime.pre_resume_pipeline(pipeline_id)

        self.runtime.set_state(node_id=pipeline_id, to_state=states.RUNNING)

        if info_list:
            self.runtime.batch_resume(process_id_list=[i.process_id for i in info_list])
            for info in info_list:
                self.runtime.execute(info.process_id, info.current_node)

        self.runtime.post_resume_pipeline(pipeline_id)

    def pause_node_appoint(self, node_id: str):
        """
        预约暂停节点

        :param node_id: 节点 ID
        :type node_id: str
        """
        node = self.runtime.get_node(node_id)

        if node.type == NodeType.SubProcess:
            raise InvalidOperationError("can not use pause_node_appoint api for {}".format(node.type))

        self.runtime.pre_pause_node(node_id)

        self.runtime.set_state(node_id=node_id, to_state=states.SUSPENDED)

        self.runtime.post_pause_node(node_id)

    def resume_node_appoint(self, node_id: str):
        """
        继续由于节点暂停被阻塞的流程的执行

        :param node_id: 节点 ID
        :type node_id: str
        """
        node = self.runtime.get_node(node_id)

        if node.type == NodeType.SubProcess:
            raise InvalidOperationError("can not use pause_node_appoint api for {}".format(node.type))

        self.runtime.pre_resume_node(node_id)

        info_list = self.runtime.get_suspended_process_info(node_id)

        self.runtime.set_state(node_id=node_id, to_state=states.READY)

        # found process suspended by node suspend
        for info in info_list:
            self.runtime.resume(process_id=info.process_id)
            self.runtime.execute(info.process_id, info.current_node)

        self.runtime.post_resume_node(node_id)

    def retry_node(self, node_id: str, data: Optional[dict] = None):
        """
        重试节点

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 重试时使用的输入数据, defaults to None
        :type data: Optional[dict], optional
        """
        node = self.runtime.get_node(node_id)

        if not node.can_retry:
            raise InvalidOperationError("can not retry node({}) with can_retry({})".format(node_id, node.can))

        state = self.runtime.get_state(node_id)

        process_id = self._ensure_state_is_fail_and_retrun_process_id(state)

        self.runtime.pre_retry_node(node_id, data)

        if data is not None:
            self.runtime.set_data_inputs(
                node_id,
                {k: DataInput(need_render=True, value=v) for k, v in data.items()},
            )

        self._add_history(node_id, state)

        self.runtime.set_state(
            node_id=node_id,
            to_state=states.READY,
            is_retry=True,
            refresh_version=True,
            clear_started_time=True,
            clear_archived_time=True,
        )

        self.runtime.execute(process_id, node_id)

        self.runtime.post_retry_node(node_id, data)

    def skip_node(self, node_id: str):
        """
        跳过失败的节点继续执行

        :param node_id: 节点 ID
        :type node_id: str
        :raises InvalidOperationError: [description]
        :raises InvalidOperationError: [description]
        """
        node = self.runtime.get_node(node_id)

        if not node.can_skip:
            raise InvalidOperationError("can not skip this node")

        if node.type not in self.PURE_SKIP_ENABLE_NODE_TYPE:
            raise InvalidOperationError("can not use skip_node api for {}".format(node.type))

        state = self.runtime.get_state(node_id)

        process_id = self._ensure_state_is_fail_and_retrun_process_id(state)

        self.runtime.pre_skip_node(node_id)

        # pure skip node type only has 1 next node
        next_node_id = node.target_nodes[0]

        self._add_history(node_id, state)

        self.runtime.set_state(
            node_id=node_id,
            to_state=states.FINISHED,
            is_skip=True,
            refresh_version=True,
            set_archive_time=True,
        )

        # 跳过节点时不再做节点输出提取到上下文的操作
        # 因为节点失败的位置未知，可能提取出来的变量是无法预知的，会导致不可预知的行为
        self.runtime.execute(process_id, next_node_id)

        self.runtime.post_skip_node(node_id)

    def skip_exclusive_gateway(self, node_id: str, flow_id: str):
        """
        跳过执行失败的分支网关继续执行

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_id: 需要继续执行的流 ID
        :type flow_id: str
        :raises InvalidOperationError: [description]
        """
        node = self.runtime.get_node(node_id)

        if node.type != NodeType.ExclusiveGateway:
            raise InvalidOperationError("{} is not exclusive gateway, actual: {}".format(node_id, node.type.value))

        next_node_id = node.targets[flow_id]

        state = self.runtime.get_state(node_id)

        process_id = self._ensure_state_is_fail_and_retrun_process_id(state)

        self.runtime.pre_skip_exclusive_gateway(node_id, flow_id)

        self._add_history(node_id, state)

        self.runtime.set_state(
            node_id=node_id,
            to_state=states.FINISHED,
            is_skip=True,
            refresh_version=True,
            set_archive_time=True,
        )

        self.runtime.execute(process_id, next_node_id)

        self.runtime.post_skip_exclusive_gateway(node_id, flow_id)

    def forced_fail_activity(self, node_id: str, ex_data: str):
        """
        强制失败某个 Activity

        :param node_id: 节点 ID
        :type node_id: str
        :param ex_data: 强制失败时写入节点有慈航数据的信息
        :type ex_data: str
        :raises InvalidOperationError: [description]
        :raises InvalidOperationError: [description]
        """
        node = self.runtime.get_node(node_id)

        if node.type != NodeType.ServiceActivity:
            raise InvalidOperationError("{} is not activity, actual: {}".format(node_id, node.type.value))

        state = self.runtime.get_state(node_id)

        if state.name != states.RUNNING:
            raise InvalidOperationError("{} state is not RUNNING, actual: {}".format(node_id, state.name))

        process_id = self.runtime.get_process_id_with_current_node_id(node_id)

        if not process_id:
            raise InvalidOperationError("can not find process with current node id: {}".format(node_id))

        self.runtime.pre_forced_fail_activity(node_id, ex_data)

        outputs = self.runtime.get_execution_data_outputs(node_id)

        outputs["ex_data"] = ex_data
        outputs["_forced_failed"] = True

        old_ver = state.version
        new_ver = self.runtime.set_state(
            node_id=node_id,
            to_state=states.FAILED,
            refresh_version=True,
            set_archive_time=True,
        )

        self.runtime.set_execution_data_outputs(node_id, outputs)

        self.runtime.kill(process_id)

        self.runtime.post_forced_fail_activity(node_id, ex_data, old_ver, new_ver)

    def callback(self, node_id: str, version: str, data: dict):
        """
        回调某个节点

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 回调执行版本
        :type version: str
        :param data: 回调数据
        :type data: dict
        :raises InvalidOperationError: [description]
        :raises InvalidOperationError: [description]
        :raises InvalidOperationError: [description]
        :raises InvalidOperationError: [description]
        """

        process_id = self.runtime.get_sleep_process_with_current_node_id(node_id)

        if not process_id:
            raise InvalidOperationError("can not find process with current node id: {}".format(node_id))

        state = self.runtime.get_state(node_id)

        schedule = self.runtime.get_schedule_with_node_and_version(node_id, version)

        if state.version != version:
            self.runtime.expire_schedule(schedule.id)
            raise InvalidOperationError("node version {} not exist".format(version))

        if schedule.finished:
            raise InvalidOperationError("scheudle is already finished")

        if schedule.expired:
            raise InvalidOperationError("scheudle is already expired")

        self.runtime.pre_callback(node_id, version, data)

        data_id = self.runtime.set_callback_data(node_id, state.version, data)

        self.runtime.schedule(process_id, node_id, schedule.id, data_id)

        self.runtime.post_callback(node_id, version, data)

    # engine event
    @setup_gauge(ENGINE_RUNNING_PROCESSES)
    @setup_histogram(ENGINE_PROCESS_RUNNING_TIME)
    def execute(self, process_id: int, node_id: str):
        """
        在某个进程上从某个节点开始进入推进循环

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        """

        process_info = self.runtime.get_process_info(process_id)
        self.runtime.wake_up(process_id)

        current_node_id = node_id

        # 推进循环
        while True:
            # 进程心跳
            self.runtime.beat(process_id)

            # 遇到推进终点后需要尝试唤醒父进程
            if current_node_id == process_info.destination_id:
                self.runtime.die(process_id)
                wake_up_seccess = self.runtime.child_process_finish(process_info.parent_id, process_id)

                if wake_up_seccess:
                    self.runtime.execute(process_info.parent_id, process_info.destination_id)

                return

            try:

                self.runtime.set_current_node(process_id, current_node_id)

                # 冻结检测
                if self.runtime.is_frozen(process_id):
                    self.runtime.freeze(process_id)
                    return

                node_state_map = self.runtime.batch_get_state_name(process_info.pipeline_stack)

                # 检测根流程是否被撤销
                if node_state_map[process_info.root_pipeline_id] == states.REVOKED:
                    self.runtime.die(process_id)
                    logger.info(
                        "[%s] root pipeline revoked checked at node %s",
                        process_info.root_pipeline_id,
                        current_node_id,
                    )
                    return

                # 检测流程栈中是否有被暂停的流程
                for pid in process_info.pipeline_stack:
                    if node_state_map[pid] == states.SUSPENDED:
                        logger.info(
                            "[%s] process %s suspended by subprocess %s",
                            process_info.root_pipeline_id,
                            process_id,
                            pid,
                        )
                        self.runtime.suspend(process_id, pid)
                        return

                node = self.runtime.get_node(current_node_id)
                node_state = self.runtime.get_state_or_none(current_node_id)
                loop = 1
                reset_mark_bit = False

                if node_state:
                    rerun_limit = self.runtime.node_rerun_limit(process_info.root_pipeline_id, current_node_id)
                    # 重入次数超过限制
                    if (
                        node_state.name == states.FINISHED
                        and node.type != NodeType.SubProcess
                        and node_state.loop > rerun_limit
                    ):
                        exec_outputs = self.runtime.get_execution_data_outputs(current_node_id)
                        exec_outputs["ex_data"] = "node execution exceed rerun limit {}".format(rerun_limit)

                        self.runtime.set_execution_data_outputs(current_node_id, exec_outputs)
                        self.runtime.set_state(
                            node_id=current_node_id,
                            to_state=states.FAILED,
                            set_archive_time=True,
                        )
                        self.runtime.sleep(process_id)

                        return

                    # 检测节点是否被预约暂停
                    if node_state.name == states.SUSPENDED:
                        # 预约暂停的节点在预约时获取不到 root_id 和 parent_id，故在此进行设置
                        self.runtime.set_state_root_and_parent(
                            node_id=current_node_id,
                            root_id=process_info.root_pipeline_id,
                            parent_id=process_info.top_pipeline_id,
                        )
                        self.runtime.suspend(process_id, current_node_id)
                        logger.info(
                            "[%s] process %s suspended by node %s",
                            process_info.root_pipeline_id,
                            process_id,
                            current_node_id,
                        )
                        return

                    # 设置状态前检测
                    if node_state.name not in states.INVERTED_TRANSITION[states.RUNNING]:
                        self.runtime.sleep(process_id)
                        return

                    if node_state.name == states.FINISHED:
                        loop = node_state.loop + 1
                        reset_mark_bit = True

                    # 重入前记录历史
                    if (
                        node.type in {NodeType.SubProcess, NodeType.ServiceActivity}
                        and node_state.name == states.FINISHED
                    ):
                        self._add_history(node_id=current_node_id, state=node_state)

                version = self.runtime.set_state(
                    node_id=current_node_id,
                    to_state=states.RUNNING,
                    loop=loop,
                    root_id=process_info.root_pipeline_id,
                    parent_id=process_info.top_pipeline_id,
                    set_started_time=True,
                    reset_skip=reset_mark_bit,
                    reset_retry=reset_mark_bit,
                    reset_error_ignored=reset_mark_bit,
                    refresh_version=reset_mark_bit,
                )

                logger.info(
                    "[%s] before execute %s(%s) state: %s",
                    process_info.root_pipeline_id,
                    node.__class__.__name__,
                    current_node_id,
                    node_state,
                )
                handler = HandlerFactory.get_handler(node, self.runtime)
                set_node_info(CurrentNodeInfo(node_id=current_node_id, version=version, loop=loop))
                type_label = self._get_metrics_node_type(node)
                execute_start = time.time()
                execute_result = handler.execute(process_info, loop, version)
                ENGINE_NODE_EXECUTE_TIME.labels(type_label).observe(time.time() - execute_start)

                # 进程是否要进入睡眠
                if execute_result.should_sleep:
                    self.runtime.sleep(process_id)

                # 节点是否准备好进入调度
                if execute_result.schedule_ready:
                    schedule = self.runtime.set_schedule(
                        process_id=process_id,
                        node_id=current_node_id,
                        version=version,
                        schedule_type=execute_result.schedule_type,
                    )
                    if execute_result.schedule_type == ScheduleType.POLL:
                        self.runtime.schedule(process_id, current_node_id, schedule.id)
                # 是否有待调度的子进程
                elif execute_result.dispatch_processes:
                    children = [d.process_id for d in execute_result.dispatch_processes]
                    logger.info(
                        "[%s] %s dispatch %s children: %s",
                        process_info.root_pipeline_id,
                        process_info.top_pipeline_id,
                        len(execute_result.dispatch_processes),
                        execute_result.dispatch_processes,
                    )
                    self.runtime.join(process_id, children)
                    for d in execute_result.dispatch_processes:
                        self.runtime.execute(d.process_id, d.node_id)

                if execute_result.should_die:
                    self.runtime.die(process_id)

                if execute_result.should_sleep or execute_result.should_die:
                    return

                current_node_id = execute_result.next_node_id
            except Exception as e:
                ex_data = traceback.format_exc()
                logger.warning(
                    "[%s]execute exception catch at node(%s): %s",
                    process_info.root_pipeline_id,
                    current_node_id,
                    ex_data,
                )

                if isinstance(e, StateVersionNotMatchError):
                    logger.warning(
                        "[%s]execute exception catch StateVersionNotMatchError at node(%s): %s",
                        process_info.root_pipeline_id,
                        current_node_id,
                        ex_data,
                    )
                    return

                self.runtime.set_state(
                    node_id=current_node_id,
                    to_state=states.FAILED,
                    root_id=process_info.root_pipeline_id,
                    parent_id=process_info.top_pipeline_id,
                    set_started_time=True,
                    set_archive_time=True,
                )

                outputs = self.runtime.get_execution_data_outputs(current_node_id)
                outputs["ex_data"] = ex_data
                self.runtime.set_execution_data_outputs(current_node_id, outputs)

                self.runtime.sleep(process_info.process_id)

                return

    @setup_gauge(ENGINE_RUNNING_SCHEDULES)
    @setup_histogram(ENGINE_SCHEDULE_RUNNING_TIME)
    def schedule(
        self,
        process_id: int,
        node_id: str,
        schedule_id: str,
        callback_data_id: Optional[int] = None,
    ):
        """
        在某个进程上开始某个节点的调度

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param schedule_id: 调度对象 ID
        :type schedule_id: str
        :param callback_data_id: 回调数据 ID, defaults to None
        :type callback_data_id: Optional[int], optional
        """
        try:
            process_info = self.runtime.get_process_info(process_id)

            # try to get lock
            lock_get = self.runtime.apply_schedule_lock(schedule_id)

            if not lock_get:
                try_after = random.randint(1, 5)
                logger.info(
                    "[%s] schedule(%s) lock %s with data %s fetch fail, try after %s",
                    process_info.root_pipeline_id,
                    node_id,
                    schedule_id,
                    callback_data_id,
                    try_after,
                )
                self.runtime.set_next_schedule(
                    process_id=process_id,
                    node_id=node_id,
                    schedule_id=schedule_id,
                    callback_data_id=callback_data_id,
                    schedule_after=try_after,
                )
                return

            with self._schedule_lock_keeper(schedule_id):
                # 进程心跳
                self.runtime.beat(process_id)

                schedule = self.runtime.get_schedule(schedule_id)

                # schedule alredy finished
                if schedule.finished:
                    return

                # 检查 schedule 是否过期
                state = self.runtime.get_state(node_id)
                if state.version != schedule.version:
                    logger.info(
                        "[%s] schedule(%s) %s with version %s expired, current version: %s",
                        process_info.root_pipeline_id,
                        schedule_id,
                        node_id,
                        schedule.version,
                        state.version,
                    )
                    self.runtime.expire_schedule(schedule_id)
                    return

                # 检查节点状态是否合法
                if state.name != states.RUNNING:
                    logger.info(
                        "[%s] schedule(%s) %s state is not running: %s",
                        process_info.root_pipeline_id,
                        schedule_id,
                        node_id,
                        state.name,
                    )
                    self.runtime.expire_schedule(schedule_id)
                    return

                # engine context prepare
                set_node_info(CurrentNodeInfo(node_id=node_id, version=state.version, loop=state.loop))

                # fetch callback data
                callback_data = None
                if callback_data_id:
                    callback_data = self.runtime.get_callback_data(callback_data_id)

                # fetch node info and start schedule
                node = self.runtime.get_node(node_id)
                handler = HandlerFactory.get_handler(node, self.runtime)
                type_label = self._get_metrics_node_type(node)

                logger.info(
                    "[%s] before schedule node %s with data %s",
                    process_info.root_pipeline_id,
                    node,
                    callback_data,
                )
                schedule_start = time.time()
                schedule_result = handler.schedule(process_info, state.loop, schedule, callback_data)
                ENGINE_NODE_SCHEDULE_TIME.labels(type_label).observe(time.time() - schedule_start)

                if schedule_result.has_next_schedule:
                    self.runtime.set_next_schedule(
                        process_info.process_id,
                        node_id,
                        schedule_id,
                        schedule_result.schedule_after,
                    )

                if schedule_result.schedule_done:
                    self.runtime.finish_schedule(schedule_id)
                    self.runtime.execute(process_id, schedule_result.next_node_id)
        except Exception as e:
            ex_data = traceback.format_exc()
            logger.warning(
                "[%s]schedule exception catch at node(%s): %s",
                process_info.root_pipeline_id,
                node_id,
                ex_data,
            )

            if isinstance(e, StateVersionNotMatchError):
                logger.exception(
                    "[%s]schedule exception catch StateVersionNotMatchError at node(%s): %s",
                    process_info.root_pipeline_id,
                    node_id,
                    ex_data,
                )
                return

            self.runtime.set_state(node_id=node_id, to_state=states.FAILED, set_archive_time=True)

            outputs = self.runtime.get_execution_data_outputs(node_id)
            outputs["ex_data"] = ex_data
            self.runtime.set_execution_data_outputs(node_id, outputs)
            self.runtime.release_schedule_lock(schedule_id)

    # help method
    @contextmanager
    def _schedule_lock_keeper(self, schedule_id: int):
        yield
        self.runtime.release_schedule_lock(schedule_id)

    def _add_history(
        self,
        node_id: str,
        state: Optional[State] = None,
        exec_data: Optional[ExecutionData] = None,
    ) -> int:
        if not state:
            state = self.runtime.get_state(node_id)

        if not exec_data:
            exec_data = self.runtime.get_execution_data(node_id)

        return self.runtime.add_history(
            node_id=node_id,
            started_time=state.started_time,
            archived_time=state.archived_time,
            loop=state.loop,
            skip=state.skip,
            retry=state.retry,
            version=state.version,
            inputs=exec_data.inputs,
            outputs=exec_data.outputs,
        )

    def _ensure_state_is_fail_and_retrun_process_id(self, state: State) -> str:
        if state.name != states.FAILED:
            raise InvalidOperationError("{} state is not FAILED, actual {}".format(state.node_id, state.name))

        process_id = self.runtime.get_sleep_process_with_current_node_id(state.node_id)

        if not process_id:
            raise InvalidOperationError("can not find sleep process with current node id: {}".format(state.node_id))

        return process_id

    def _get_metrics_node_type(self, node: Node) -> str:
        if node.type != NodeType.ServiceActivity:
            return get_lower_case_name(node.type.value)

        return "{}_{}".format(node.code, node.version)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from bamboo_engine import states
from bamboo_engine.config import Settings
from bamboo_engine.eri import ProcessInfo, ContextValue, ContextValueType, NodeType
from bamboo_engine.handler import register_handler, NodeHandler, ExecuteResult
from bamboo_engine.context import Context

logger = logging.getLogger("bamboo_engine")


@register_handler(NodeType.EmptyEndEvent)
class EmptyEndEventHandler(NodeHandler):
    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :return: 执行结果
        :rtype: ExecuteResult
        """
        root_pipeline_id = process_info.root_pipeline_id
        pipeline_id = process_info.pipeline_stack.pop()
        root_pipeline_finished = len(process_info.pipeline_stack) == 0

        root_pipeline_inputs = self._get_plain_inputs(process_info.root_pipeline_id)
        if not root_pipeline_finished:
            subproc_state = self.runtime.get_state(pipeline_id)

        # write pipeline data
        context_outputs = self.runtime.get_context_outputs(pipeline_id)
        logger.info(
            "[%s] %s context outputs: %s",
            root_pipeline_id,
            pipeline_id,
            context_outputs,
        )

        context_values = self.runtime.get_context_values(
            pipeline_id=pipeline_id, keys=context_outputs
        )
        logger.info(
            "[%s] %s context values: %s",
            root_pipeline_id,
            pipeline_id,
            context_values,
        )

        context = Context(self.runtime, context_values, root_pipeline_inputs)
        hydrated_context = context.hydrate(deformat=False)
        logger.info(
            "[%s] %s hydrated context: %s",
            root_pipeline_id,
            pipeline_id,
            hydrated_context,
        )

        outputs = {}
        for key in context_outputs:
            outputs[key] = hydrated_context.get(key, key)
        if not root_pipeline_finished:
            outputs[self.LOOP_KEY] = subproc_state.loop + Settings.RERUN_INDEX_OFFSET
        self.runtime.set_execution_data_outputs(node_id=pipeline_id, outputs=outputs)

        self.runtime.set_state(
            node_id=self.node.id, to_state=states.FINISHED, set_archive_time=True
        )

        self.runtime.set_state(
            node_id=pipeline_id, to_state=states.FINISHED, set_archive_time=True
        )

        # root pipeline finish
        if root_pipeline_finished:
            return ExecuteResult(
                should_sleep=False,
                schedule_ready=False,
                schedule_type=None,
                schedule_after=-1,
                dispatch_processes=[],
                next_node_id=None,
                should_die=True,
            )

        # subprocess finish
        subprocess = self.runtime.get_node(pipeline_id)
        self.runtime.set_pipeline_stack(
            process_info.process_id, process_info.pipeline_stack
        )

        # extract subprocess outputs to parent context
        subprocess_outputs = self.runtime.get_data_outputs(pipeline_id)
        context.extract_outputs(
            pipeline_id=process_info.pipeline_stack[-1],
            data_outputs=subprocess_outputs,
            execution_data_outputs=outputs,
        )

        return ExecuteResult(
            should_sleep=False,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=[],
            next_node_id=subprocess.target_nodes[0],
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

"""
封装模板处理，渲染逻辑的相关模块
"""

import copy
import re
import logging

from typing import Any, List, Set

from mako.template import Template as MakoTemplate
from mako import lexer, codegen
from mako.exceptions import MakoException

from bamboo_engine.utils.mako_utils.checker import check_mako_template_safety
from bamboo_engine.utils.mako_utils.exceptions import ForbiddenMakoTemplateException
from bamboo_engine.utils import mako_safety
from bamboo_engine.utils.string import deformat_var_key

from . import sandbox


logger = logging.getLogger("root")
# find mako template(format is ${xxx}，and ${}# not in xxx, # may raise memory error)
TEMPLATE_PATTERN = re.compile(r"\${[^${}#]+}")


class Template:
    def __init__(self, data: Any):
        self.data = data

    def get_reference(self, deformat=False) -> Set[str]:
        """
        获取当前数据中模板所引用的所有标志符

        :return: 标志符列表
        :rtype: List[str]
        """

        reference = []
        templates = self.get_templates()
        for tpl in templates:
            reference += self._get_template_reference(tpl)
        reference = set(reference)
        if not deformat:
            reference = {"${%s}" % r for r in reference}

        return reference

    def get_templates(self) -> List[str]:
        """
        获取当前数据中所有的模板片段

        :return: 模板片段列表
        :rtype: List[str]
        """
        templates = []
        data = self.data
        if isinstance(data, str):
            templates += self._get_string_templates(data)
        if isinstance(data, (list, tuple)):
            for item in data:
                templates += Template(item).get_templates()
        if isinstance(data, dict):
            for value in list(data.values()):
                templates += Template(value).get_templates()
        return list(set(templates))

    def render(self, context: dict) -> Any:
        """
        渲染当前模板

        :param context: 模板渲染上下文
        :type context: dict
        :return: 模板渲染后的数据
        :rtype: Any
        """
        data = self.data
        if isinstance(data, str):
            return self._render_string(data, context)
        if isinstance(data, list):
            ldata = [""] * len(data)
            for index, item in enumerate(data):
                ldata[index] = Template(copy.deepcopy(item)).render(context)
            return ldata
        if isinstance(data, tuple):
            ldata = [""] * len(data)
            for index, item in enumerate(data):
                ldata[index] = Template(copy.deepcopy(item)).render(context)
            return tuple(ldata)
        if isinstance(data, dict):
            for key, value in list(data.items()):
                data[key] = Template(copy.deepcopy(value)).render(context)
            return data
        return data

    def _get_string_templates(self, string) -> List[str]:
        return list(set(TEMPLATE_PATTERN.findall(string)))

    def _get_template_reference(self, template: str) -> List[str]:
        lex = lexer.Lexer(template)

        try:
            node = lex.parse()
        except MakoException as e:
            logger.warning(
                "pipeline get template[{}] reference error[{}]".format(template, e)
            )
            return []

        # Dummy compiler. _Identifiers class requires one
        # but only interested in the reserved_names field
        def compiler():
            return None

        compiler.reserved_names = set()
        identifiers = codegen._Identifiers(compiler, node)

        return list(identifiers.undeclared)

    def _render_string(self, string: str, context: dict) -> str:
        """
        使用特定上下文渲染指定模板

        :param string: 模板
        :type string: str
        :param context: 上下文
        :type context: dict
        :return: 渲染后的模板
        :rtype: str
        """
        if not isinstance(string, str):
            return string
        templates = self._get_string_templates(string)

        # TODO keep render return object, here only process simple situation
        if (
            len(templates) == 1
            and templates[0] == string
            and deformat_var_key(string) in context
        ):
            return context[deformat_var_key(string)]

        for tpl in templates:
            try:
                check_mako_template_safety(
                    tpl,
                    mako_safety.SingleLineNodeVisitor(),
                    mako_safety.SingleLinCodeExtractor(),
                )
            except ForbiddenMakoTemplateException as e:
                logger.warning("forbidden template: {}, exception: {}".format(tpl, e))
                continue
            except Exception:
                logger.exception("{} safety check error.".format(tpl))
                continue
            resolved = Template._render_template(tpl, context)
            string = string.replace(tpl, resolved)
        return string

    @staticmethod
    def _render_template(template: str, context: dict) -> Any:
        """
        使用特定上下文渲染指定模板

        :param template: 模板
        :type template: Any
        :param context: 上下文
        :type context: dict
        :raises TypeError: [description]
        :return: [description]
        :rtype: str
        """
        data = {}
        data.update(sandbox.get())
        data.update(context)
        if not isinstance(template, str):
            raise TypeError(
                "constant resolve error, template[%s] is not a string" % template
            )
        try:
            tm = MakoTemplate(template)
        except (MakoException, SyntaxError) as e:
            logger.error("pipeline resolve template[{}] error[{}]".format(template, e))
            return template
        try:
            resolved = tm.render_unicode(**data)
        except Exception as e:
            logger.warning(
                "constant content({}) is invalid, data({}), error: {}".format(
                    template, data, e
                )
            )
            return template
        else:
            return resolved



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback
from typing import Optional

from bamboo_engine import states
from bamboo_engine.config import Settings

from bamboo_engine.context import Context
from bamboo_engine.template import Template
from bamboo_engine.eri import (
    ProcessInfo,
    ContextValue,
    ContextValueType,
    ExecutionData,
    CallbackData,
    ScheduleType,
    NodeType,
    Schedule,
)
from bamboo_engine.handler import (
    register_handler,
    NodeHandler,
    ExecuteResult,
    ScheduleResult,
)

logger = logging.getLogger("bamboo_engine")


@register_handler(NodeType.ServiceActivity)
class ServiceActivityHandler(NodeHandler):
    """
    其中所有 set_state 调用都会传入 state version 来确保能够在用户强制失败节点后放弃后续无效的任务执行
    """

    def execute(
        self, process_info: ProcessInfo, loop: int, version: str
    ) -> ExecuteResult:
        """
        节点的 execute 处理逻辑

        :param runtime: 引擎运行时实例
        :type runtime: EngineRuntimeInterface
        :param process_info: 进程信息
        :type process_id: ProcessInfo
        :return: 执行结果
        :rtype: ExecuteResult
        """
        top_pipeline_id = process_info.top_pipeline_id
        root_pipeline_id = process_info.root_pipeline_id

        data = self.runtime.get_data(self.node.id)
        root_pipeline_inputs = self._get_plain_inputs(process_info.root_pipeline_id)
        need_render_inputs = data.need_render_inputs()
        render_escape_inputs = data.render_escape_inputs()

        logger.info(
            "[%s] %s activity execute data: %s, root inputs: %s",
            root_pipeline_id,
            self.node.id,
            data,
            root_pipeline_inputs,
        )

        # resolve inputs context references
        inputs_refs = set(Template(need_render_inputs).get_reference())
        logger.info(
            "[%s] %s activity original refs: %s",
            root_pipeline_id,
            self.node.id,
            inputs_refs,
        )

        additional_refs = self.runtime.get_context_key_references(
            pipeline_id=top_pipeline_id, keys=inputs_refs
        )
        inputs_refs = inputs_refs.union(additional_refs)
        logger.info(
            "[%s] %s activity final refs: %s",
            root_pipeline_id,
            self.node.id,
            inputs_refs,
        )

        # prepare context
        context_values = self.runtime.get_context_values(
            pipeline_id=top_pipeline_id, keys=inputs_refs
        )

        # pre extract loop outputs
        loop_value = loop + Settings.RERUN_INDEX_OFFSET
        need_render_inputs[self.LOOP_KEY] = loop_value
        if self.LOOP_KEY in data.outputs:
            loop_output_key = data.outputs[self.LOOP_KEY]
            context_values.append(
                ContextValue(
                    key=loop_output_key, type=ContextValueType.PLAIN, value=loop_value
                )
            )
        logger.info(
            "[%s] %s activity context values: %s",
            root_pipeline_id,
            self.node.id,
            context_values,
        )

        context = Context(self.runtime, context_values, root_pipeline_inputs)
        hydrated_context = context.hydrate(deformat=True)
        logger.info(
            "[%s] %s actvity hydrated context: %s",
            root_pipeline_id,
            self.node.id,
            hydrated_context,
        )

        # resolve inputs
        execute_inputs = Template(need_render_inputs).render(hydrated_context)
        execute_inputs.update(render_escape_inputs)

        # data prepare
        service_data = ExecutionData(inputs=execute_inputs, outputs={})
        root_pipeline_data = ExecutionData(inputs=root_pipeline_inputs, outputs={})

        # execute
        service = self.runtime.get_service(
            code=self.node.code, version=self.node.version
        )
        service.setup_runtime_attributes(
            id=self.node.id,
            version=version,
            top_pipeline_id=top_pipeline_id,
            root_pipeline_id=root_pipeline_id,
            loop=loop,
        )

        # start monitor
        monitoring = False
        if self.node.timeout is not None:
            monitoring = True
            self.runtime.start_timeout_monitor(
                process_id=process_info.process_id,
                node_id=self.node.id,
                version=version,
                timeout=self.node.timeout,
            )

        # pre_execute and excute
        logger.debug(
            "[%s] %s service data before execute: %s",
            self.node.id,
            root_pipeline_id,
            service_data,
        )
        logger.debug(
            "[%s] %s root pipeline data before execute: %s",
            self.node.id,
            root_pipeline_id,
            root_pipeline_data,
        )
        execute_success = False
        try:
            service.pre_execute(
                data=service_data, root_pipeline_data=root_pipeline_data
            )
            execute_success = service.execute(
                data=service_data, root_pipeline_data=root_pipeline_data
            )
        except Exception:
            ex_data = traceback.format_exc()
            service_data.outputs.ex_data = ex_data
            logger.warning(
                "[%s]service execute fail: %s", process_info.root_pipeline_id, ex_data
            )
        logger.debug(
            "[%s] service data after execute: %s", root_pipeline_id, service_data
        )
        service_data.outputs._result = execute_success

        # execute success
        if execute_success:

            need_schedule = service.need_schedule()
            next_node_id = None

            if not need_schedule:
                if monitoring:
                    self.runtime.stop_timeout_monitor(
                        process_id=process_info.process_id,
                        node_id=self.node.id,
                        version=version,
                        timeout=self.node.timeout,
                    )

                self.runtime.set_state(
                    node_id=self.node.id,
                    version=version,
                    to_state=states.FINISHED,
                    set_archive_time=True,
                )

                context.extract_outputs(
                    pipeline_id=top_pipeline_id,
                    data_outputs=data.outputs,
                    execution_data_outputs=service_data.outputs,
                )
                next_node_id = self.node.target_nodes[0]

            self.runtime.set_execution_data(node_id=self.node.id, data=service_data)

            return ExecuteResult(
                should_sleep=need_schedule,
                schedule_ready=need_schedule,
                schedule_type=service.schedule_type(),
                schedule_after=service.schedule_after(
                    schedule=None,
                    data=service_data,
                    root_pipeline_data=root_pipeline_data,
                ),
                dispatch_processes=[],
                next_node_id=next_node_id,
            )

        # pre_execute failed or execute failed
        if monitoring:
            self.runtime.stop_timeout_monitor(
                process_id=process_info.process_id,
                node_id=self.node.id,
                version=version,
                timeout=self.node.timeout,
            )

        if not self.node.error_ignorable:
            self.runtime.set_state(
                node_id=self.node.id,
                version=version,
                to_state=states.FAILED,
                set_archive_time=True,
            )

            self.runtime.set_execution_data(node_id=self.node.id, data=service_data)

            context.extract_outputs(
                pipeline_id=top_pipeline_id,
                data_outputs=data.outputs,
                execution_data_outputs=service_data.outputs,
            )

            return ExecuteResult(
                should_sleep=True,
                schedule_ready=False,
                schedule_type=None,
                schedule_after=-1,
                dispatch_processes=[],
                next_node_id=None,
            )

        # pre_execute failed or execute failed and error ignore
        self.runtime.set_state(
            node_id=self.node.id,
            version=version,
            to_state=states.FINISHED,
            set_archive_time=True,
            error_ignored=True,
        )

        self.runtime.set_execution_data(node_id=self.node.id, data=service_data)

        context.extract_outputs(
            pipeline_id=top_pipeline_id,
            data_outputs=data.outputs,
            execution_data_outputs=service_data.outputs,
        )

        return ExecuteResult(
            should_sleep=False,
            schedule_ready=False,
            schedule_type=None,
            schedule_after=-1,
            dispatch_processes=[],
            next_node_id=self.node.target_nodes[0],
        )

    def _finish_schedule(
        self,
        process_info: ProcessInfo,
        schedule: Schedule,
        data_outputs: dict,
        execution_data: ExecutionData,
        error_ignored: bool,
        root_pipeline_inputs: dict,
    ) -> ScheduleResult:
        if self.node.timeout is not None:
            self.runtime.stop_timeout_monitor(
                process_id=process_info.process_id,
                node_id=self.node.id,
                version=schedule.version,
                timeout=self.node.timeout,
            )

        self.runtime.set_state(
            node_id=self.node.id,
            version=schedule.version,
            to_state=states.FINISHED,
            set_archive_time=True,
            error_ignored=error_ignored,
        )

        context = Context(self.runtime, [], root_pipeline_inputs)
        context.extract_outputs(
            pipeline_id=process_info.top_pipeline_id,
            data_outputs=data_outputs,
            execution_data_outputs=execution_data.outputs,
        )

        return ScheduleResult(
            has_next_schedule=False,
            schedule_after=-1,
            schedule_done=True,
            next_node_id=self.node.target_nodes[0],
        )

    def schedule(
        self,
        process_info: ProcessInfo,
        loop: int,
        schedule: Schedule,
        callback_data: Optional[CallbackData] = None,
    ) -> ScheduleResult:
        """
        节点的 schedule 处理逻辑

        :param process_id: 进程 ID
        :type process_id: int
        :param schedule: Schedule 实例
        :type schedule: Schedule
        :param callback_data: 回调数据, defaults to None
        :type callback_data: Optional[CallbackData], optional
        :return: 调度结果
        :rtype: ScheduleResult
        """
        # data prepare
        top_pipeline_id = process_info.top_pipeline_id
        root_pipeline_id = process_info.root_pipeline_id

        data_outputs = self.runtime.get_data_outputs(self.node.id)
        service_data = self.runtime.get_execution_data(self.node.id)

        root_pipeline_inputs = self._get_plain_inputs(root_pipeline_id)
        root_pipeline_data = ExecutionData(inputs=root_pipeline_inputs, outputs={})
        logger.info(
            "[%s] %s activity schedule data: %s, root inputs: %s",
            root_pipeline_id,
            self.node.id,
            service_data,
            root_pipeline_inputs,
        )

        # schedule
        service = self.runtime.get_service(
            code=self.node.code, version=self.node.version
        )
        service.setup_runtime_attributes(
            id=self.node.id,
            version=schedule.version,
            top_pipeline_id=top_pipeline_id,
            root_pipeline_id=root_pipeline_id,
            loop=loop,
        )

        schedule_success = False
        schedule.times += 1
        try:
            schedule_success = service.schedule(
                schedule=schedule,
                data=service_data,
                root_pipeline_data=root_pipeline_data,
                callback_data=callback_data,
            )
        except Exception:
            service_data.outputs.ex_data = traceback.format_exc()

        service_data.outputs._result = schedule_success
        self.runtime.add_schedule_times(schedule.id)
        self.runtime.set_execution_data(node_id=self.node.id, data=service_data)

        monitoring = self.node.timeout is not None
        schedule_type = service.schedule_type()

        # schedule success
        if schedule_success:
            if schedule_type == ScheduleType.CALLBACK:
                return self._finish_schedule(
                    process_info=process_info,
                    schedule=schedule,
                    data_outputs=data_outputs,
                    execution_data=service_data,
                    error_ignored=False,
                    root_pipeline_inputs=root_pipeline_inputs,
                )
            else:
                is_schedule_done = service.is_schedule_done()

                # poll or multi-callback finished
                if is_schedule_done:
                    return self._finish_schedule(
                        process_info=process_info,
                        schedule=schedule,
                        data_outputs=data_outputs,
                        execution_data=service_data,
                        error_ignored=False,
                        root_pipeline_inputs=root_pipeline_inputs,
                    )

                has_next_schedule = schedule_type == ScheduleType.POLL
                return ScheduleResult(
                    has_next_schedule=has_next_schedule,
                    schedule_after=service.schedule_after(
                        schedule=schedule,
                        data=service_data,
                        root_pipeline_data=root_pipeline_data,
                    ),
                    schedule_done=False,
                    next_node_id=None,
                )

        if monitoring:
            self.runtime.stop_timeout_monitor(
                process_id=process_info.process_id,
                node_id=self.node.id,
                version=schedule.version,
                timeout=self.node.timeout,
            )

        # schedule fail
        if not self.node.error_ignorable:
            self.runtime.set_state(
                node_id=self.node.id,
                version=schedule.version,
                to_state=states.FAILED,
                set_archive_time=True,
            )

            context = Context(self.runtime, [], root_pipeline_inputs)
            context.extract_outputs(
                pipeline_id=process_info.top_pipeline_id,
                data_outputs=data_outputs,
                execution_data_outputs=service_data.outputs,
            )

            return ScheduleResult(
                has_next_schedule=False,
                schedule_after=-1,
                schedule_done=False,
                next_node_id=None,
            )

        # schedule fail and error ignore
        return self._finish_schedule(
            process_info=process_info,
            schedule=schedule,
            data_outputs=data_outputs,
            execution_data=service_data,
            error_ignored=True,
            root_pipeline_inputs=root_pipeline_inputs,
        )
